# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM wall command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Wall:
    """The command definition for the Arch workbench's gui tool, Arch Wall.

    A tool for creating Arch walls.

    Create a wall from the object selected by the user. If no objects are
    selected, enter an interactive mode to create a wall using selected points
    to create a base.

    Find documentation on the end user usage of Arch Wall here:
    https://wiki.freecad.org/Arch_Wall
    """

    def GetResources(self):
        """Returns a dictionary with the visual aspects of the Arch Wall tool."""

        return {
            "Pixmap": "Arch_Wall",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Wall", "Wall"),
            "Accel": "W, A",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Wall",
                "Creates a wall object from scratch or from a selected object (wire, face or solid)",
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        """Executed when Arch Wall is called.

        Creates a wall from the object selected by the user. If no objects are
        selected, enters an interactive mode to create a wall using selected
        points to create a base.
        """

        import Draft
        import WorkingPlane
        from draftutils import params
        import draftguitools.gui_trackers as DraftTrackers

        self.doc = FreeCAD.ActiveDocument
        self.Align = ["Center", "Left", "Right"][params.get_param_arch("WallAlignment")]
        self.MultiMat = None
        self.Length = None
        self.lengthValue = 0
        self.featureName = "Wall"
        self.Width = params.get_param_arch("WallWidth")
        self.Height = params.get_param_arch("WallHeight")
        self.Offset = params.get_param_arch("WallOffset")
        self.JOIN_WALLS_SKETCHES = params.get_param_arch("joinWallSketches")
        self.AUTOJOIN = params.get_param_arch("autoJoinWalls")
        sel = FreeCADGui.Selection.getSelectionEx()
        self.existing = []
        self.wp = None

        if sel:
            # automatic mode
            if Draft.getType(sel[0].Object) != "Wall":
                self.doc.openTransaction(translate("Arch", "Create Wall"))
                FreeCADGui.addModule("Arch")
                for selobj in sel:
                    if (
                        Draft.getType(selobj.Object) == "Space"
                        and selobj.HasSubObjects
                        and "Face" in selobj.SubElementNames[0]
                    ):
                        idx = int(selobj.SubElementNames[0][4:])
                        FreeCADGui.doCommand(
                            "obj = Arch.makeWall(FreeCAD.ActiveDocument."
                            + selobj.Object.Name
                            + ",face="
                            + str(idx)
                            + ")"
                        )
                    else:
                        FreeCADGui.doCommand(
                            "obj = Arch.makeWall(FreeCAD.ActiveDocument." + selobj.Object.Name + ")"
                        )
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand("Draft.autogroup(obj)")
                self.doc.commitTransaction()
                self.doc.recompute()
                return

        # interactive mode

        FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
        self.points = []
        self.wp = WorkingPlane.get_working_plane()
        self.tracker = DraftTrackers.boxTracker()
        FreeCADGui.Snapper.getPoint(
            callback=self.getPoint,
            extradlg=self.taskbox(),
            title=translate("Arch", "First point of wall"),
        )
        FreeCADGui.draftToolBar.continueCmd.show()

    def getPoint(self, point=None, obj=None):
        """Callback for clicks during interactive mode.

        When method _CommandWall.Activated() has entered the interactive mode,
        this callback runs when the user clicks.

        Parameters
        ----------
        point: <class 'Base.Vector'>
            The point the user has selected.
        obj: <Part::Feature>, optional
            The object the user's cursor snapped to, if any.
        """

        import Draft
        import ArchWall
        from draftutils import gui_utils

        if obj:
            if Draft.getType(obj) == "Wall":
                if not obj in self.existing:
                    self.existing.append(obj)
        if point is None:
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            self.tracker.finalize()
            return
        self.points.append(point)
        if len(self.points) == 1:
            self.tracker.width(self.Width)
            self.tracker.height(self.Height)
            self.tracker.on()
            FreeCADGui.Snapper.getPoint(
                last=self.points[0],
                callback=self.getPoint,
                movecallback=self.update,
                extradlg=self.taskbox(),
                title=translate("Arch", "Next point"),
                mode="line",
            )

        elif len(self.points) == 2:
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            self.tracker.off()

            self.doc.openTransaction(translate("Arch", "Create Wall"))

            # Some code in gui_utils.autogroup requires a wall shape to determine
            # the target group. We therefore need to create a wall first.
            self.addDefault()
            wall = self.doc.Objects[-1]
            wallGrp = wall.getParentGroup()

            if (
                (self.JOIN_WALLS_SKETCHES or self.AUTOJOIN)
                and self.existing
                and self.existing[-1].getParentGroup() == wallGrp
            ):
                oldWall = self.existing[-1]
                if self.JOIN_WALLS_SKETCHES and ArchWall.areSameWallTypes([wall, oldWall]):
                    FreeCADGui.doCommand(
                        "Arch.joinWalls([wall, doc."
                        + oldWall.Name
                        + "], "
                        + "delete=True, deletebase=True)"
                    )
                elif self.AUTOJOIN:
                    if wallGrp is not None:
                        FreeCADGui.doCommand("wall.getParentGroup().removeObject(wall)")
                    FreeCADGui.doCommand("Arch.addComponents(wall, doc." + oldWall.Name + ")")

            self.doc.commitTransaction()
            self.doc.recompute()
            # gui_utils.end_all_events()  # Causes a crash on Linux.
            self.tracker.finalize()
            if FreeCADGui.draftToolBar.continueMode:
                self.Activated()

    def addDefault(self):
        """Create a wall using a line segment, with all parameters as the default.

        Used solely by _CommandWall.getPoint() when the interactive mode has
        selected two points.
        """
        from draftutils import params

        sta = self.wp.get_local_coords(self.points[0])
        end = self.wp.get_local_coords(self.points[1])

        FreeCADGui.doCommand("import Part")
        FreeCADGui.addModule("Draft")
        FreeCADGui.addModule("Arch")
        FreeCADGui.addModule("WorkingPlane")
        FreeCADGui.doCommand("doc = FreeCAD.ActiveDocument")
        FreeCADGui.doCommand("wp = WorkingPlane.get_working_plane()")
        FreeCADGui.doCommand(
            "trace = Part.LineSegment(FreeCAD." + str(sta) + ", FreeCAD." + str(end) + ")"
        )
        if params.get_param_arch("WallSketches"):
            # Use ArchSketch if SketchArch add-on is present
            try:
                import ArchSketchObject

                FreeCADGui.doCommand("import ArchSketchObject")
                FreeCADGui.doCommand("base = ArchSketchObject.makeArchSketch()")
            except:
                FreeCADGui.doCommand('base = doc.addObject("Sketcher::SketchObject", "WallTrace")')
            FreeCADGui.doCommand("base.Placement = wp.get_placement()")
            FreeCADGui.doCommand("base.addGeometry(trace)")
        else:
            FreeCADGui.doCommand("base = Draft.make_line(trace)")
            # The created line should not stay selected as this causes an issue in continue mode.
            # Two walls would then be created based on the same line.
            FreeCADGui.Selection.clearSelection()
            FreeCADGui.doCommand("base.Placement = wp.get_placement()")
            FreeCADGui.doCommand("doc.recompute()")
        FreeCADGui.doCommand(
            "wall = Arch.makeWall(base, width="
            + str(self.Width)
            + ", height="
            + str(self.Height)
            + ', align="'
            + str(self.Align)
            + '")'
        )
        FreeCADGui.doCommand("wall.Normal = wp.axis")
        if self.MultiMat:
            FreeCADGui.doCommand("wall.Material = doc." + self.MultiMat.Name)
        FreeCADGui.doCommand(
            "doc.recompute()"
        )  # required as some autogroup code requires the wall shape
        FreeCADGui.doCommand("Draft.autogroup(wall)")

    def update(self, point, info):
        # info parameter is not used but needed for compatibility with the snapper

        """Callback for the mouse moving during the interactive mode.

        Update the active dialog box to show the coordinates of the location of
        the cursor. Also show the length the line would take, if the user
        selected that point.

        Parameters
        ----------
        point: <class 'Base.Vector'>
            The point the cursor is currently at, or has snapped to.
        """

        import DraftVecUtils

        if FreeCADGui.Control.activeDialog():
            b = self.points[0]
            n = self.wp.axis
            bv = point.sub(b)
            dv = bv.cross(n)
            dv = DraftVecUtils.scaleTo(dv, self.Width / 2)
            if self.Align == "Center":
                self.tracker.update([b, point])
            elif self.Align == "Left":
                self.tracker.update([b.add(dv), point.add(dv)])
            else:
                dv = dv.negative()
                self.tracker.update([b.add(dv), point.add(dv)])
            if self.Length:
                self.Length.setText(
                    FreeCAD.Units.Quantity(bv.Length, FreeCAD.Units.Length).UserString
                )

    def taskbox(self):
        """Set up a simple gui widget for the interactive mode."""

        from PySide import QtCore, QtGui
        import Draft
        from draftutils import params

        w = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        w.setWindowTitle(translate("Arch", "Wall options"))
        grid = QtGui.QGridLayout(w)

        # Wall presets input
        comboWallPresets = QtGui.QComboBox()
        comboWallPresets.addItem(translate("Arch", "Wall Presets"))
        comboWallPresets.setToolTip(
            translate(
                "Arch",
                "This list shows all the MultiMaterials objects of this document. Create some to define wall types.",
            )
        )
        self.multimats = []
        self.MultiMat = None
        for o in self.doc.Objects:
            if Draft.getType(o) == "MultiMaterial":
                self.multimats.append(o)
                comboWallPresets.addItem(o.Label)
        if hasattr(FreeCAD, "LastArchMultiMaterial"):
            for i, o in enumerate(self.multimats):
                if o.Name == FreeCAD.LastArchMultiMaterial:
                    comboWallPresets.setCurrentIndex(i + 1)
                    self.MultiMat = o
        grid.addWidget(comboWallPresets, 0, 0, 1, 2)

        # Wall length input
        labelLength = QtGui.QLabel(translate("Arch", "Length"))
        self.Length = ui.createWidget("Gui::InputField")
        self.Length.setText("0.00 mm")
        grid.addWidget(labelLength, 1, 0, 1, 1)
        grid.addWidget(self.Length, 1, 1, 1, 1)

        # Wall width input
        labelWidth = QtGui.QLabel(translate("Arch", "Width"))
        inputWidth = ui.createWidget("Gui::InputField")
        inputWidth.setText(FreeCAD.Units.Quantity(self.Width, FreeCAD.Units.Length).UserString)
        grid.addWidget(labelWidth, 2, 0, 1, 1)
        grid.addWidget(inputWidth, 2, 1, 1, 1)

        # Wall height input
        labelHeight = QtGui.QLabel(translate("Arch", "Height"))
        inputHeight = ui.createWidget("Gui::InputField")
        inputHeight.setText(FreeCAD.Units.Quantity(self.Height, FreeCAD.Units.Length).UserString)
        grid.addWidget(labelHeight, 3, 0, 1, 1)
        grid.addWidget(inputHeight, 3, 1, 1, 1)

        # Wall alignment input
        labelAlignment = QtGui.QLabel(translate("Arch", "Alignment"))
        comboAlignment = QtGui.QComboBox()
        items = [translate("Arch", "Center"), translate("Arch", "Left"), translate("Arch", "Right")]
        comboAlignment.addItems(items)
        comboAlignment.setCurrentIndex(["Center", "Left", "Right"].index(self.Align))
        grid.addWidget(labelAlignment, 4, 0, 1, 1)
        grid.addWidget(comboAlignment, 4, 1, 1, 1)

        # Wall offset input
        labelOffset = QtGui.QLabel(translate("Arch", "Offset"))
        inputOffset = ui.createWidget("Gui::InputField")
        inputOffset.setText(FreeCAD.Units.Quantity(self.Offset, FreeCAD.Units.Length).UserString)
        grid.addWidget(labelOffset, 5, 0, 1, 1)
        grid.addWidget(inputOffset, 5, 1, 1, 1)

        # Wall "use sketches" checkbox
        labelUseSketches = QtGui.QLabel(translate("Arch", "Use sketches"))
        checkboxUseSketches = QtGui.QCheckBox()
        checkboxUseSketches.setObjectName("UseSketches")
        checkboxUseSketches.setLayoutDirection(QtCore.Qt.RightToLeft)
        labelUseSketches.setBuddy(checkboxUseSketches)
        checkboxUseSketches.setChecked(params.get_param_arch("WallSketches"))
        grid.addWidget(labelUseSketches, 6, 0, 1, 1)
        grid.addWidget(checkboxUseSketches, 6, 1, 1, 1)

        # Enable/disable inputOffset based on inputAlignment
        def updateOffsetState(index):
            alignment = ["Center", "Left", "Right"][index]
            inputOffset.setEnabled(alignment in ["Left", "Right"])

        # Connect the signal
        comboAlignment.currentIndexChanged.connect(updateOffsetState)

        # Initialize the state of inputOffset
        updateOffsetState(comboAlignment.currentIndex())

        # Connect the signals to the slots for value changes
        self.Length.valueChanged.connect(self.setLength)
        inputWidth.valueChanged.connect(self.setWidth)
        inputHeight.valueChanged.connect(self.setHeight)
        comboAlignment.currentIndexChanged.connect(self.setAlign)
        inputOffset.valueChanged.connect(self.setOffset)
        if hasattr(checkboxUseSketches, "checkStateChanged"):  # Qt version >= 6.7.0
            checkboxUseSketches.checkStateChanged.connect(self.setUseSketch)
        else:  # Qt version < 6.7.0
            checkboxUseSketches.stateChanged.connect(self.setUseSketch)
        comboWallPresets.currentIndexChanged.connect(self.setMat)

        # Define the workflow of the input fields:
        # Pressing Enter will cycle through Length, Width, and
        # finally on Height the wall will be created
        self.Length.returnPressed.connect(inputWidth.setFocus)
        self.Length.returnPressed.connect(inputWidth.selectAll)
        inputWidth.returnPressed.connect(inputHeight.setFocus)
        inputWidth.returnPressed.connect(inputHeight.selectAll)
        inputHeight.returnPressed.connect(self.createFromGUI)
        inputOffset.returnPressed.connect(self.createFromGUI)

        return w

    def setMat(self, d):
        """Simple callback for the interactive mode gui widget to set material."""

        if d == 0:
            self.MultiMat = None
            del FreeCAD.LastArchMultiMaterial
        elif d <= len(self.multimats):
            self.MultiMat = self.multimats[d - 1]
            FreeCAD.LastArchMultiMaterial = self.MultiMat.Name

    def setLength(self, d):
        """Simple callback for the interactive mode gui widget to set length."""

        if isinstance(d, FreeCAD.Units.Quantity):
            d = d.Value
        self.lengthValue = d

    def setWidth(self, d):
        """Simple callback for the interactive mode gui widget to set width."""

        from draftutils import params

        if isinstance(d, FreeCAD.Units.Quantity):
            d = d.Value
        self.Width = d
        self.tracker.width(d)
        params.set_param_arch("WallWidth", d)

    def setHeight(self, d):
        """Simple callback for the interactive mode gui widget to set height."""

        from draftutils import params

        if isinstance(d, FreeCAD.Units.Quantity):
            d = d.Value
        self.Height = d
        self.tracker.height(d)
        params.set_param_arch("WallHeight", d)

    def setAlign(self, i):
        """Simple callback for the interactive mode gui widget to set alignment."""

        from draftutils import params

        self.Align = ["Center", "Left", "Right"][i]
        params.set_param_arch("WallAlignment", i)

    def setOffset(self, d):
        """Simple callback for the interactive mode GUI widget to set offset."""

        from draftutils import params

        if isinstance(d, FreeCAD.Units.Quantity):
            d = d.Value
        self.Offset = d
        params.set_param_arch("WallOffset", d)

    def setUseSketch(self, i):
        """Simple callback to set if walls should based on sketches."""

        from draftutils import params

        params.set_param_arch("WallSketches", bool(getattr(i, "value", i)))

    def createFromGUI(self):
        """Callback to create wall by using the _CommandWall.taskbox()"""

        self.doc.openTransaction(translate("Arch", "Create Wall"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.doCommand(
            "wall = Arch.makeWall(length="
            + str(self.lengthValue)
            + ",width="
            + str(self.Width)
            + ",height="
            + str(self.Height)
            + ',align="'
            + str(self.Align)
            + '")'
        )
        if self.MultiMat:
            FreeCADGui.doCommand("wall.Material = FreeCAD.ActiveDocument." + self.MultiMat.Name)
        self.doc.commitTransaction()
        self.doc.recompute()
        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.escape()


FreeCADGui.addCommand("Arch_Wall", Arch_Wall())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Equipment"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchEquipment
#  \ingroup ARCH
#  \brief The Equipment object and tools
#
#  This module provides tools to build equipment objects.
#  Equipment is used to represent furniture and all kinds of electrical
#  or hydraulic appliances in a building

import FreeCAD
import ArchComponent
import DraftVecUtils

if FreeCAD.GuiUp:
    from PySide import QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _Equipment(ArchComponent.Component):
    "The Equipment object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Equipment"
        self.setProperties(obj)
        from ArchIFC import IfcTypes

        if "Furniture" in IfcTypes:
            # IfcFurniture is new in IFC4
            obj.IfcType = "Furniture"
        elif "Furnishing Element" in IfcTypes:
            # IFC2x3 does know a IfcFurnishingElement
            obj.IfcType = "Furnishing Element"
        else:
            obj.IfcType = "Building Element Proxy"
        # Add features in the SketchArch External Add-on, if present
        self.addSketchArchFeatures(obj)

    def addSketchArchFeatures(self, obj, linkObj=None, mode=None):
        """
        To add features in the SketchArch External Add-on, if present (https://github.com/paullee0/FreeCAD_SketchArch)
        -  import ArchSketchObject module, and
        -  set properties that are common to ArchObjects (including Links) and ArchSketch
           to support the additional features

        To install SketchArch External Add-on, see https://github.com/paullee0/FreeCAD_SketchArch#iv-install
        """

        try:
            import ArchSketchObject

            ArchSketchObject.ArchSketch.setPropertiesLinkCommon(self, obj, linkObj, mode)
        except:
            pass

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Model" in pl:
            obj.addProperty(
                "App::PropertyString",
                "Model",
                "Equipment",
                QT_TRANSLATE_NOOP("App::Property", "The model description of this equipment"),
                locked=True,
            )
        if not "ProductURL" in pl:
            obj.addProperty(
                "App::PropertyString",
                "ProductURL",
                "Equipment",
                QT_TRANSLATE_NOOP("App::Property", "The URL of the product page of this equipment"),
                locked=True,
            )
        if not "StandardCode" in pl:
            obj.addProperty(
                "App::PropertyString",
                "StandardCode",
                "Equipment",
                QT_TRANSLATE_NOOP("App::Property", "A standard code (MasterFormat, OmniClass,…)"),
                locked=True,
            )
        if not "SnapPoints" in pl:
            obj.addProperty(
                "App::PropertyVectorList",
                "SnapPoints",
                "Equipment",
                QT_TRANSLATE_NOOP("App::Property", "Additional snap points for this equipment"),
                locked=True,
            )
        if not "EquipmentPower" in pl:
            obj.addProperty(
                "App::PropertyFloat",
                "EquipmentPower",
                "Equipment",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The electric power needed by this equipment in Watts"
                ),
                locked=True,
            )
        obj.setEditorMode("VerticalArea", 2)
        obj.setEditorMode("HorizontalArea", 2)
        obj.setEditorMode("PerimeterLength", 2)

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

        # Add features in the SketchArch External Add-on, if present
        self.addSketchArchFeatures(obj)

    def loads(self, state):

        self.Type = "Equipment"

    def onChanged(self, obj, prop):

        self.hideSubobjects(obj, prop)
        ArchComponent.Component.onChanged(self, obj, prop)

    def execute(self, obj):

        if self.clone(obj):
            return
        if not self.ensureBase(obj):
            return

        pl = obj.Placement
        if obj.Base:
            base = None
            if hasattr(obj.Base, "Shape"):
                base = obj.Base.Shape.copy()
                base = self.processSubShapes(obj, base, pl)
                self.applyShape(obj, base, pl, allowinvalid=False, allownosolid=True)

        # Execute features in the SketchArch External Add-on, if present
        self.executeSketchArchFeatures(obj)

    def executeSketchArchFeatures(self, obj, linkObj=None, index=None, linkElement=None):
        """
        To execute features in the SketchArch External Add-on  (https://github.com/paullee0/FreeCAD_SketchArch)
        -  import ArchSketchObject module, and
        -  execute features that are common to ArchObjects (including Links) and ArchSketch

        To install SketchArch External Add-on, see https://github.com/paullee0/FreeCAD_SketchArch#iv-install
        """

        # To execute features in SketchArch External Add-on, if present
        try:
            import ArchSketchObject

            # Execute SketchArch Feature - Intuitive Automatic Placement for Arch Windows/Doors, Equipment etc.
            # see https://forum.freecad.org/viewtopic.php?f=23&t=50802
            ArchSketchObject.updateAttachmentOffset(obj, linkObj)
        except:
            pass

    def appLinkExecute(self, obj, linkObj, index, linkElement):
        """
        Default Link Execute method() -
        See https://forum.freecad.org/viewtopic.php?f=22&t=42184&start=10#p361124
        @realthunder added support to Links to run Linked Scripted Object's methods()
        """

        # Add features in the SketchArch External Add-on, if present
        self.addSketchArchFeatures(obj, linkObj)

        # Execute features in the SketchArch External Add-on, if present
        self.executeSketchArchFeatures(obj, linkObj)

    def computeAreas(self, obj):
        return


class _ViewProviderEquipment(ArchComponent.ViewProviderComponent):
    "A View Provider for the Equipment object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)

    def getIcon(self):

        import Arch_rc

        if hasattr(self, "Object"):
            if hasattr(self.Object, "CloneOf"):
                if self.Object.CloneOf:
                    return ":/icons/Arch_Equipment_Clone.svg"
        return ":/icons/Arch_Equipment_Tree.svg"

    def attach(self, vobj):

        self.Object = vobj.Object
        from pivy import coin

        sep = coin.SoSeparator()
        self.coords = coin.SoCoordinate3()
        sep.addChild(self.coords)
        self.coords.point.deleteValues(0)
        symbol = coin.SoMarkerSet()
        symbol.markerIndex = FreeCADGui.getMarkerIndex("", 5)
        sep.addChild(symbol)
        rn = vobj.RootNode
        rn.addChild(sep)
        ArchComponent.ViewProviderComponent.attach(self, vobj)

    def updateData(self, obj, prop):

        if prop == "SnapPoints":
            if obj.SnapPoints:
                self.coords.point.setNum(len(obj.SnapPoints))
                self.coords.point.setValues([[p.x, p.y, p.z] for p in obj.SnapPoints])
            else:
                self.coords.point.deleteValues(0)
        else:
            ArchComponent.ViewProviderComponent.updateData(self, obj, prop)



# ***************************************************************************
# *   Copyright (c) 2001,2002 Juergen Riegel <juergen.riegel@web.de>        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# FreeCAD Part module
#
# Part design module

# import FreeCAD modules
import FreeCAD, FreeCADGui

# import the App Test module
import TestApp  # Test as Module name not possible
import sys


# ---------------------------------------------------------------------------
# define the Commands of the Test Application module
# ---------------------------------------------------------------------------
class TestCmd:
    """Opens a Qt dialog with all inserted unit tests"""

    def Activated(self):
        import QtUnitGui

        tests = FreeCAD.__unit_test__

        QtUnitGui.addTest("TestApp.All")
        QtUnitGui.setTest("TestApp.All")
        for test in tests:
            QtUnitGui.addTest(test)

    def GetResources(self):
        return {
            "MenuText": "Self-test...",
            "ToolTip": "Runs a self-test to check if the application works properly",
        }


class TestAllCmd:
    "Test all commando object"

    def Activated(self):
        import QtUnitGui

        QtUnitGui.addTest("TestApp.All")
        QtUnitGui.setTest("TestApp.All")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test all",
            "ToolTip": "Runs all tests at once (can take very long!)",
        }


class TestDocCmd:
    "Document test commando object"

    def Activated(self):
        import QtUnitGui

        QtUnitGui.addTest("Document")
        QtUnitGui.setTest("Document")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test Document",
            "ToolTip": "Test the document (creation, save, load and destruction)",
        }


class TestBaseCmd:
    "Base test commando object"

    def Activated(self):
        import QtUnitGui

        QtUnitGui.addTest("BaseTests")
        QtUnitGui.setTest("BaseTests")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test base",
            "ToolTip": "Test the basic functions of FreeCAD",
        }


class TestAllTextCmd:
    "Test all commando object"

    def Activated(self):
        import unittest, TestApp

        unittest.TextTestRunner(stream=sys.stdout, verbosity=2).run(
            unittest.defaultTestLoader.loadTestsFromName("TestApp.All")
        )

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test all",
            "ToolTip": "Runs all tests at once (can take very long!)",
        }


class TestDocTextCmd:
    "Document test commando object"

    def Activated(self):
        TestApp.TestText("Document")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test Document",
            "ToolTip": "Test the document (creation, save, load and destruction)",
        }


class TestBaseTextCmd:
    "Base test commando object"

    def Activated(self):
        TestApp.TestText("BaseTests")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test base",
            "ToolTip": "Test the basic functions of FreeCAD",
        }


class TestWorkbenchCmd:
    "Workbench test"

    def Activated(self):
        i = 0
        while i < 20:
            FreeCADGui.activateWorkbench("MeshWorkbench")
            FreeCADGui.updateGui()
            FreeCADGui.activateWorkbench("NoneWorkbench")
            FreeCADGui.updateGui()
            FreeCADGui.activateWorkbench("PartWorkbench")
            FreeCADGui.updateGui()
            print(i)
            i = i + 1
        FreeCADGui.activateWorkbench("TestWorkbench")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Test workbench",
            "ToolTip": "Test the switching of workbenches in FreeCAD",
        }


class TestCreateMenuCmd:
    "Base test commando object"

    def Activated(self):
        TestApp.TestText("Menu.MenuCreateCases")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Add menu",
            "ToolTip": "Test the menu stuff of FreeCAD",
        }


class TestDeleteMenuCmd:
    "Base test commando object"

    def Activated(self):
        TestApp.TestText("Menu.MenuDeleteCases")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Remove menu",
            "ToolTip": "Test the menu stuff of FreeCAD",
        }


class TestInsertFeatureCmd:
    "Base test commando object"

    def Activated(self):
        if FreeCAD.activeDocument() is not None:
            FreeCAD.activeDocument().addObject("App::FeatureTest")
        else:
            FreeCAD.PrintMessage("No active document.\n")

    def GetResources(self):
        return {
            "Pixmap": "Std_Tool1",
            "MenuText": "Insert a TestFeature",
            "ToolTip": "Insert a TestFeature in the active Document",
        }


# ---------------------------------------------------------------------------
# Adds the commands to the FreeCAD command manager
# ---------------------------------------------------------------------------
FreeCADGui.addCommand("Test_Test", TestCmd())
FreeCADGui.addCommand("Test_TestAllText", TestAllTextCmd())
FreeCADGui.addCommand("Test_TestDocText", TestDocTextCmd())
FreeCADGui.addCommand("Test_TestBaseText", TestBaseTextCmd())
FreeCADGui.addCommand("Test_TestAll", TestAllCmd())
FreeCADGui.addCommand("Test_TestDoc", TestDocCmd())
FreeCADGui.addCommand("Test_TestBase", TestBaseCmd())
FreeCADGui.addCommand("Test_TestWork", TestWorkbenchCmd())
FreeCADGui.addCommand("Test_TestCreateMenu", TestCreateMenuCmd())
FreeCADGui.addCommand("Test_TestDeleteMenu", TestDeleteMenuCmd())
FreeCADGui.addCommand("Test_InsertFeature", TestInsertFeatureCmd())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create BezCurve objects."""
## @package make_bezcurve
# \ingroup draftmake
# \brief Provides functions to create BezCurve objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftutils.translate import translate
from draftobjects.bezcurve import BezCurve

if App.GuiUp:
    from draftviewproviders.view_bezcurve import ViewProviderBezCurve


def make_bezcurve(pointslist, closed=False, placement=None, face=None, support=None, degree=None):
    """make_bezcurve(pointslist, [closed], [placement])

    Creates a Bezier Curve object from the given list of vectors.

    Parameters
    ----------
    pointlist : [Base.Vector]
        List of points to create the polyline.
        Instead of a pointslist, you can also pass a Part Wire.
        TODO: Change the name so!

    closed : bool
        If closed is True or first and last points are identical,
        the created BSpline will be closed.

    placement : Base.Placement
        If a placement is given, it is used.

    face : Bool
        If face is False, the rectangle is shown as a wireframe,
        otherwise as a face.

    support :
        TODO: Describe

    degree : int
        Degree of the BezCurve
    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return
    if not isinstance(pointslist, list):
        nlist = []
        for v in pointslist.Vertexes:
            nlist.append(v.Point)
        pointslist = nlist
    if placement:
        utils.type_check([(placement, App.Placement)], "make_bezcurve")
    if len(pointslist) == 2:
        fname = "Line"
    else:
        fname = "BezCurve"
    obj = App.ActiveDocument.addObject("Part::FeaturePython", fname)
    obj.addExtension("Part::AttachExtensionPython")
    BezCurve(obj)
    obj.Points = pointslist
    if degree:
        obj.Degree = degree
    else:
        import Part

        obj.Degree = min((len(pointslist) - (1 * (not closed))), Part.BezierCurve().MaxDegree)
    obj.Closed = closed
    obj.AttachmentSupport = support
    if face is not None:
        obj.MakeFace = face
    obj.Proxy.resetcontinuity(obj)
    if placement:
        obj.Placement = placement
    if App.GuiUp:
        ViewProviderBezCurve(obj.ViewObject)
        #        if not face: obj.ViewObject.DisplayMode = "Wireframe"
        #        obj.ViewObject.DisplayMode = "Wireframe"
        gui_utils.format_object(obj)
        gui_utils.select(obj)

    return obj


makeBezCurve = make_bezcurve

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create dimension objects.

The objects can be simple linear dimensions that measure between two arbitrary
points, or linear dimensions linked to an edge.
It can also be radius or diameter dimensions that measure circles
and circular arcs.
And it can also be an angular dimension measuring the angle between
two straight lines.
"""
## @package gui_dimensions
# \ingroup draftguitools
# \brief Provides GUI tools to create dimension objects.

import math
import lazy_loader.lazy_loader as lz
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import DraftVecUtils
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftguitools.gui_trackers as trackers
import draftutils.gui_utils as gui_utils

from draftutils.translate import translate
from draftutils.messages import _toolmsg, _msg

DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False

## \addtogroup draftguitools
# @{


class Dimension(gui_base_original.Creator):
    """Gui command for the Dimension tool.

    This includes at the moment linear, radial, diametrical,
    and angular dimensions depending on the selected object
    and the modifier key (ALT) used.

    Maybe in the future each type can be in its own class,
    and they can inherit their basic properties from a parent class.
    """

    def __init__(self):
        super().__init__()
        self.max = 2
        self.chain = None  # Last chain's leg in ChainMode
        self.contMode = None
        self.dir = None
        self.featureName = "Dimension"

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Dimension",
            "Accel": "D, I",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Dimension", "Dimension"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Dimension",
                "Creates a linear dimension for a straight edge, a circular edge, or 2 picked points, or an angular dimension for 2 straight edges",
            ),
        }

    def Activated(self, dir_vec=None):
        """Execute when the command is called."""
        if self.chain and not self.contMode:
            self.finish()
        else:
            super().Activated(name=self.featureName)
            if self.ui:
                self.ui.pointUi(title=translate("draft", self.featureName), icon="Draft_Dimension")
                self.ui.continueCmd.show()
                self.ui.chainedModeCmd.show()
                self.ui.selectButton.show()
                self.altdown = False
                self.call = self.view.addEventCallback("SoEvent", self.action)
                self.dimtrack = trackers.dimTracker()
                self.arctrack = trackers.arcTracker()
                self.link = None
                self.dir = dir_vec
                self.edges = []
                self.angles = []
                self.angledata = None
                self.indices = []
                self.center = None
                self.arcmode = False
                self.point1 = None
                self.point2 = None
                self.proj_point1 = None
                self.proj_point2 = None
                self.force = None
                self.info = None
                self.selectmode = False
                self.set_selection()
                _toolmsg(translate("draft", "Pick first point"))

    def set_selection(self):
        """Fill the nodes according to the selected geometry."""
        sel = Gui.Selection.getSelectionEx()
        if (
            len(sel) == 1
            and len(sel[0].SubElementNames) == 1
            and "Edge" in sel[0].SubElementNames[0]
        ):
            # The selection is just a single `Edge`
            sel_object = sel[0]
            edge = sel_object.SubObjects[0]

            # `n` is the edge number starting from 0 not from 1.
            # The reason is lists in Python start from 0, although
            # in the object's `Shape`, they start from 1
            n = int(sel_object.SubElementNames[0].lstrip("Edge")) - 1
            self.indices.append(n)

            if DraftGeomUtils.geomType(edge) == "Line":
                self.node.extend([edge.Vertexes[0].Point, edge.Vertexes[1].Point])

                # Iterate over the vertices of the parent `Object`;
                # when the vertices match those of the selected `edge`
                # save the index of vertex in the parent object
                v1 = None
                v2 = None
                for i, v in enumerate(sel_object.Object.Shape.Vertexes):
                    if v.Point == edge.Vertexes[0].Point:
                        v1 = i
                    if v.Point == edge.Vertexes[1].Point:
                        v2 = i

                if v1 is not None and v2 is not None:  # note that v1 or v2 can be zero
                    self.link = [sel_object.Object, v1, v2]
            elif DraftGeomUtils.geomType(edge) == "Circle":
                self.node.extend([edge.Curve.Center, edge.Vertexes[0].Point])
                self.edges = [edge]
                self.arcmode = "diameter"
                self.link = [sel_object.Object, n]

    def finish(self, cont=False):
        """Terminate the operation."""
        self.end_callbacks(self.call)
        self.chain = None
        self.contMode = None
        self.dir = None
        if self.ui:
            self.dimtrack.finalize()
            self.arctrack.finalize()
        super().finish()

    def angle_dimension_normal(self, edge1, edge2):
        rot = App.Rotation(
            DraftGeomUtils.vec(edge1), DraftGeomUtils.vec(edge2), self.wp.axis, "XYZ"
        )
        norm = rot.multVec(App.Vector(0, 0, 1))
        vnorm = gui_utils.get_3d_view().getViewDirection()
        if vnorm.getAngle(norm) < math.pi / 2:
            norm = norm.negative()
        return norm

    def create_angle_dimension(self):
        """Create an angular dimension from a center and two angles."""
        ang1 = math.degrees(self.angledata[1])
        ang2 = math.degrees(self.angledata[0])
        norm = self.angle_dimension_normal(self.edges[0], self.edges[1])

        _cmd = "Draft.make_angular_dimension"
        _cmd += "("
        _cmd += "center=" + DraftVecUtils.toString(self.center) + ", "
        _cmd += "angles="
        _cmd += "["
        _cmd += str(ang1) + ", "
        _cmd += str(ang2)
        _cmd += "], "
        _cmd += "dim_line=" + DraftVecUtils.toString(self.node[-1]) + ", "
        _cmd += "normal=" + DraftVecUtils.toString(norm)
        _cmd += ")"
        _cmd_list = [
            "_dim_ = " + _cmd,
            "Draft.autogroup(_dim_)",
            "FreeCAD.ActiveDocument.recompute()",
        ]
        self.commit(translate("draft", "Create Dimension"), _cmd_list)

    def create_linear_dimension(self):
        """Create a simple linear dimension, not linked to an edge."""
        _cmd = "Draft.make_linear_dimension"
        _cmd += "("
        _cmd += DraftVecUtils.toString(self.node[0]) + ", "
        _cmd += DraftVecUtils.toString(self.node[1]) + ", "
        _cmd += "dim_line=" + DraftVecUtils.toString(self.node[2])
        _cmd += ")"
        _cmd_list = [
            "_dim_ = " + _cmd,
            "Draft.autogroup(_dim_)",
            "FreeCAD.ActiveDocument.recompute()",
        ]
        self.commit(translate("draft", "Create Dimension"), _cmd_list)

    def create_linear_dimension_obj(self, direction=None):
        """Create a linear dimension linked to an edge.

        The `link` attribute has indices of vertices as they appear
        in the list `Shape.Vertexes`, so they start as zero 0.

        The `LinearDimension` class, created by `make_linear_dimension_obj`,
        considers the vertices of a `Shape` which are numbered to start
        with 1, that is, `Vertex1`.
        Therefore the value in `link` has to be incremented by 1.
        """
        _cmd = "Draft.make_linear_dimension_obj"
        _cmd += "("
        _cmd += "FreeCAD.ActiveDocument." + self.link[0].Name + ", "
        _cmd += "i1=" + str(self.link[1] + 1) + ", "
        _cmd += "i2=" + str(self.link[2] + 1) + ", "
        _cmd += "dim_line=" + DraftVecUtils.toString(self.node[2])
        _cmd += ")"
        _cmd_list = ["_dim_ = " + _cmd]

        dir_u = DraftVecUtils.toString(self.wp.u)
        dir_v = DraftVecUtils.toString(self.wp.v)
        if direction == "X":
            _cmd_list += ["_dim_.Direction = " + dir_u]
        elif direction == "Y":
            _cmd_list += ["_dim_.Direction = " + dir_v]

        _cmd_list += ["Draft.autogroup(_dim_)", "FreeCAD.ActiveDocument.recompute()"]
        self.commit(translate("draft", "Create Dimension"), _cmd_list)

    def create_radial_dimension_obj(self):
        """Create a radial dimension linked to a circular edge."""
        _cmd = "Draft.make_radial_dimension_obj"
        _cmd += "("
        _cmd += "FreeCAD.ActiveDocument." + self.link[0].Name + ", "
        _cmd += "index=" + str(self.link[1] + 1) + ", "
        _cmd += 'mode="' + str(self.arcmode) + '", '
        _cmd += "dim_line=" + DraftVecUtils.toString(self.node[2])
        _cmd += ")"
        _cmd_list = [
            "_dim_ = " + _cmd,
            "Draft.autogroup(_dim_)",
            "FreeCAD.ActiveDocument.recompute()",
        ]
        self.commit(translate("draft", "Create Dimension"), _cmd_list)

    def createObject(self):
        """Create the actual object in the current document."""
        Gui.addModule("Draft")

        if self.angledata:
            # Angle dimension, with two angles provided
            self.create_angle_dimension()
        elif self.link and not self.arcmode:
            # Linear dimension, linked to a straight edge
            if self.force == 1:
                self.create_linear_dimension_obj("Y")
            elif self.force == 2:
                self.create_linear_dimension_obj("X")
            else:
                self.create_linear_dimension_obj()
        elif self.arcmode:
            # Radius or dimeter dimension, linked to a circular edge
            self.create_radial_dimension_obj()
        else:
            # Linear dimension, not linked to any edge
            self.create_linear_dimension()

        if self.ui.chainedMode or self.ui.continueMode:
            if self.ui.chainedMode:
                self.chain = self.node[2]
            if not self.dir:
                if self.link:
                    v1 = self.link[0].Shape.Vertexes[self.link[1]].Point
                    v2 = self.link[0].Shape.Vertexes[self.link[2]].Point
                    self.dir = v2.sub(v1)
                else:
                    self.dir = self.node[1].sub(self.node[0])

            self.node = [self.node[1]]

        self.link = None

    def selectEdge(self):
        """Toggle the select mode to the opposite state."""
        self.selectmode = not self.selectmode

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            shift = gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_constrain_key())
            if self.arcmode or self.point2:
                gui_tool_utils.setMod(arg, gui_tool_utils.get_mod_constrain_key(), False)
            (self.point, ctrlPoint, self.info) = gui_tool_utils.getPoint(
                self, arg, noTracker=(len(self.node) > 0)
            )
            if (
                gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key()) or self.selectmode
            ) and (len(self.node) < 3):
                self.dimtrack.off()
                if not self.altdown:
                    self.altdown = True
                    self.ui.switchUi(True)
                    if hasattr(Gui, "Snapper"):
                        Gui.Snapper.setSelectMode(True)
                snapped = self.view.getObjectInfo((arg["Position"][0], arg["Position"][1]))
                if snapped:
                    ob = self.doc.getObject(snapped["Object"])
                    if "Edge" in snapped["Component"]:
                        num = int(snapped["Component"].lstrip("Edge")) - 1
                        ed = ob.Shape.Edges[num]
                        v1 = ed.Vertexes[0].Point
                        v2 = ed.Vertexes[-1].Point
                        self.dimtrack.update([v1, v2, self.chain])
            else:
                if self.node and (len(self.edges) < 2):
                    self.dimtrack.on()
                if len(self.edges) == 2:
                    # angular dimension
                    self.dimtrack.off()

                    vnorm = gui_utils.get_3d_view().getViewDirection()
                    anorm = self.arctrack.normal

                    # Code below taken from WorkingPlane.projectPoint:
                    cos = vnorm.dot(anorm)
                    delta_ax_proj = (self.point - self.center).dot(anorm)
                    proj = self.point - delta_ax_proj / cos * vnorm
                    self.point = proj

                    r = self.point.sub(self.center)
                    self.arctrack.setRadius(r.Length)
                    a = self.arctrack.getAngle(self.point)
                    pair = DraftGeomUtils.getBoundaryAngles(a, self.angles)
                    if not (pair[0] < a < pair[1]):
                        self.angledata = [4 * math.pi - pair[0], 2 * math.pi - pair[1]]
                    else:
                        self.angledata = [2 * math.pi - pair[0], 2 * math.pi - pair[1]]
                    self.arctrack.setStartAngle(self.angledata[0])
                    self.arctrack.setEndAngle(self.angledata[1])
                if self.altdown:
                    self.altdown = False
                    self.ui.switchUi(False)
                    if hasattr(Gui, "Snapper"):
                        Gui.Snapper.setSelectMode(False)
                if self.node and self.dir and len(self.node) < 2:
                    _p = DraftVecUtils.project(self.point.sub(self.node[0]), self.dir)
                    self.point = self.node[0].add(_p)
                if len(self.node) == 2:
                    if self.arcmode and self.edges:
                        cen = self.edges[0].Curve.Center
                        rad = self.edges[0].Curve.Radius
                        baseray = self.point.sub(cen)
                        v2 = DraftVecUtils.scaleTo(baseray, rad)
                        v1 = v2.negative()
                        if shift:
                            self.node = [cen, cen.add(v2)]
                            self.arcmode = "radius"
                        else:
                            self.node = [cen.add(v1), cen.add(v2)]
                            self.arcmode = "diameter"
                        self.dimtrack.update(self.node)
                # Draw constraint tracker line.
                if shift and (not self.arcmode):
                    if len(self.node) == 2:
                        if not self.point1:
                            self.point1 = self.node[0]
                        if not self.point2:
                            self.point2 = self.node[1]
                        # else:
                        #     self.node[1] = self.point2
                        self.set_constraint_node()
                else:
                    self.force = None
                    self.proj_point1 = None
                    self.proj_point2 = None
                    if self.point1:
                        self.node[0] = self.point1
                    if self.point2 and (len(self.node) > 1):
                        self.node[1] = self.point2
                        # self.point2 = None
                # update the dimline
                if self.node and not self.arcmode:
                    self.dimtrack.update(self.node + [self.point] + [self.chain])
            gui_tool_utils.redraw3DView()
        elif arg["Type"] == "SoMouseButtonEvent":
            if (arg["State"] == "DOWN") and (arg["Button"] == "BUTTON1"):
                if self.point:
                    self.ui.redraw()
                    if (not self.node) and (not self.support):
                        gui_tool_utils.getSupport(arg)
                    if (
                        gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key())
                        or self.selectmode
                    ) and (len(self.node) < 3):
                        # print("snapped: ",self.info)
                        if self.info:
                            ob = self.doc.getObject(self.info["Object"])
                            if "Edge" in self.info["Component"]:
                                num = int(self.info["Component"].lstrip("Edge")) - 1
                                ed = ob.Shape.Edges[num]
                                v1 = ed.Vertexes[0].Point
                                v2 = ed.Vertexes[-1].Point
                                i1 = i2 = None
                                for i in range(len(ob.Shape.Vertexes)):
                                    if v1 == ob.Shape.Vertexes[i].Point:
                                        i1 = i
                                    if v2 == ob.Shape.Vertexes[i].Point:
                                        i2 = i
                                if (i1 is not None) and (i2 is not None):
                                    self.indices.append(num)
                                    if not self.edges:
                                        # nothing snapped yet, we treat it
                                        # as a normal edge-snapped dimension
                                        self.node = [v1, v2]
                                        self.link = [ob, i1, i2]
                                        self.edges.append(ed)
                                        if DraftGeomUtils.geomType(ed) == "Circle":
                                            # snapped edge is an arc
                                            self.arcmode = "diameter"
                                            self.link = [ob, num]
                                    else:
                                        # there is already a snapped edge,
                                        # so we start angular dimension
                                        self.edges.append(ed)
                                        # self.node now has the 4 endpoints
                                        self.node.extend([v1, v2])
                                        c = DraftGeomUtils.findIntersection(
                                            self.node[0],
                                            self.node[1],
                                            self.node[2],
                                            self.node[3],
                                            True,
                                            True,
                                        )
                                        if c:
                                            # print("centers:",c)
                                            self.center = c[0]
                                            self.arctrack.setCenter(self.center)
                                            self.arctrack.normal = self.angle_dimension_normal(
                                                self.edges[0], self.edges[1]
                                            )
                                            self.arctrack.on()
                                            for e in self.edges:
                                                if (
                                                    e.Length < 0.00003
                                                ):  # Edge must be long enough for the tolerance of 0.00001mm to make sense.
                                                    _msg(translate("draft", "Edge too short!"))
                                                    self.finish()
                                                    return
                                                for i in [0, 1]:
                                                    pt = e.Vertexes[i].Point
                                                    if pt.isEqual(
                                                        self.center, 0.00001
                                                    ):  # A relatively high tolerance is required.
                                                        pt = e.Vertexes[
                                                            i - 1
                                                        ].Point  # Use the other point instead.
                                                    self.angles.append(self.arctrack.getAngle(pt))
                                            self.link = [self.link[0], ob]
                                        else:
                                            _msg(translate("draft", "Edges do not intersect!"))
                                            self.finish()
                                            return
                                self.dimtrack.on()
                    else:
                        self.node.append(self.point)
                    self.selectmode = False
                    # print("node", self.node)
                    self.dimtrack.update(self.node)
                    if len(self.node) == 2:
                        self.point2 = self.node[1]
                    if len(self.node) == 1:
                        self.dimtrack.on()
                        if self.planetrack:
                            self.planetrack.set(self.node[0])
                    elif len(self.node) == 2 and self.chain:
                        self.node.append(self.chain)
                        self.createObject()
                        if not self.chain:
                            self.finish()
                    elif len(self.node) == 3:
                        # for unlinked arc mode:
                        # if self.arcmode:
                        #     v = self.node[1].sub(self.node[0])
                        #     v.multiply(0.5)
                        #     cen = self.node[0].add(v)
                        #     self.node = [self.node[0], self.node[1], cen]
                        self.createObject()
                        if self.ui.continueMode:
                            self.contMode = True
                            self.Activated()
                        elif not self.chain:
                            self.finish()
                    elif self.angledata:
                        self.node.append(self.point)
                        self.createObject()
                        self.finish()

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        self.node.append(self.point)
        self.dimtrack.update(self.node)
        if len(self.node) == 1:
            self.dimtrack.on()
        elif len(self.node) == 3:
            self.createObject()
            if not self.chain:
                self.finish()

    def set_constraint_node(self):
        """Set constrained nodes for vertical or horizontal dimension
        by projecting on the working plane.
        """
        if not self.proj_point1 or not self.proj_point2:
            self.proj_point1 = self.wp.project_point(self.node[0])
            self.proj_point2 = self.wp.project_point(self.node[1])
            proj_u = self.wp.u.dot(self.proj_point2 - self.proj_point1)
            proj_v = self.wp.v.dot(self.proj_point2 - self.proj_point1)
            active_view = Gui.ActiveDocument.ActiveView
            cursor = active_view.getCursorPos()
            cursor_point = active_view.getPoint(cursor)
            self.point = self.wp.project_point(cursor_point)
            if not self.force:
                ref_point = self.point - (self.proj_point2 + self.proj_point1) * 1 / 2
                ref_angle = abs(ref_point.getAngle(self.wp.u))
                if (ref_angle > math.pi / 4) and (ref_angle <= 0.75 * math.pi):
                    self.force = 2
                else:
                    self.force = 1
            if self.force == 1:
                self.node[0] = self.proj_point1
                self.node[1] = self.proj_point1 + self.wp.v * proj_v
            elif self.force == 2:
                self.node[0] = self.proj_point1
                self.node[1] = self.proj_point1 + self.wp.u * proj_u


Gui.addCommand("Draft_Dimension", Dimension())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Gui.CircularHoleBase as PathCircularHoleBaseGui
import Path.Op.Helix as PathHelix
import PathGui
from PySide.QtCore import QT_TRANSLATE_NOOP

translate = FreeCAD.Qt.translate

from PySide import QtCore

__doc__ = "Helix operation page controller and command implementation."

LOGLEVEL = False

if LOGLEVEL:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.NOTICE, Path.Log.thisModule())


class TaskPanelOpPage(PathCircularHoleBaseGui.TaskPanelOpPage):
    """Page controller class for Helix operations."""

    def getForm(self):
        """getForm() ... return UI"""

        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpHelixEdit.ui")
        comboToPropertyMap = [("startSide", "StartSide"), ("cutMode", "CutMode")]

        enumTups = PathHelix.ObjectHelix.helixOpPropertyEnumerations(dataType="raw")

        self.populateCombobox(form, enumTups, comboToPropertyMap)
        return form

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        Path.Log.track()
        if obj.CutMode != str(self.form.cutMode.currentData()):
            obj.CutMode = str(self.form.cutMode.currentData())
        if obj.StartSide != str(self.form.startSide.currentData()):
            obj.StartSide = str(self.form.startSide.currentData())
        if obj.StepOver != self.form.stepOverPercent.value():
            obj.StepOver = self.form.stepOverPercent.value()
        PathGuiUtil.updateInputField(obj, "OffsetExtra", self.form.extraOffset)

        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        Path.Log.track()

        self.form.stepOverPercent.setValue(obj.StepOver)
        self.selectInComboBox(obj.CutMode, self.form.cutMode)
        self.selectInComboBox(obj.StartSide, self.form.startSide)

        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)

        self.form.extraOffset.setText(
            FreeCAD.Units.Quantity(obj.OffsetExtra.Value, FreeCAD.Units.Length).UserString
        )

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []

        signals.append(self.form.stepOverPercent.editingFinished)
        signals.append(self.form.extraOffset.editingFinished)
        signals.append(self.form.cutMode.currentIndexChanged)
        signals.append(self.form.startSide.currentIndexChanged)
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)

        return signals


Command = PathOpGui.SetupOperation(
    "Helix",
    PathHelix.Create,
    TaskPanelOpPage,
    "CAM_Helix",
    QT_TRANSLATE_NOOP("CAM_Helix", "Helix"),
    QT_TRANSLATE_NOOP("CAM_Helix", "Creates a Helical toolpath from the features of a base object"),
    PathHelix.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathHelixGui... done\n")



# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import FreeCAD
import Path
from typing import Any, Dict, List, Optional


class DetachedDocumentObject:
    """
    A lightweight class mimicking the property API of a FreeCAD DocumentObject.

    This class is used by ToolBit instances when they are not associated
    with a real FreeCAD DocumentObject, allowing properties to be stored
    and accessed in a detached state.
    """

    def __init__(self, label: str = "DetachedObject"):
        self.Label: str = label
        self.Name: str = label.replace(" ", "_")
        self.PropertiesList: List[str] = []
        self._properties: Dict[str, Any] = {}
        self._property_groups: Dict[str, Optional[str]] = {}
        self._property_types: Dict[str, Optional[str]] = {}
        self._property_docs: Dict[str, Optional[str]] = {}
        self._editor_modes: Dict[str, int] = {}
        self._property_enums: Dict[str, List[str]] = {}

    def addProperty(
        self,
        thetype: Optional[str],
        name: str,
        group: Optional[str],
        doc: Optional[str],
    ) -> None:
        """Mimics FreeCAD DocumentObject.addProperty."""
        if name not in self._properties:
            self.PropertiesList.append(name)
            self._properties[name] = None
            self._property_groups[name] = group
            self._property_types[name] = thetype
            self._property_docs[name] = doc
            if thetype in [
                "App::PropertyQuantity",
                "App::PropertyLength",
                "App::PropertyArea",
                "App::PropertyVolume",
                "App::PropertyAngle",
            ]:
                # Initialize Quantity properties with a default value
                self._properties[name] = FreeCAD.Units.Quantity(0.0)

    def removeProperty(self, name: str) -> None:
        """Removes a property from the detached object."""
        if name in self._properties:
            if name in self.PropertiesList:
                self.PropertiesList.remove(name)
            del self._properties[name]
            self._property_groups.pop(name, None)
            self._property_types.pop(name, None)
            self._property_docs.pop(name, None)
            self._editor_modes.pop(name, None)
            self._property_enums.pop(name, None)

    def getPropertyByName(self, name: str) -> Any:
        """Mimics FreeCAD DocumentObject.getPropertyByName."""
        return self._properties.get(name)

    def setPropertyByName(self, name: str, value: Any) -> None:
        """Mimics FreeCAD DocumentObject.setPropertyByName."""
        self._properties[name] = value

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Intercept attribute assignment. This is done to behave like
        FreeCAD's DocumentObject, which may have any property assigned,
        pre-defined or not.
        Without this, code linters report an error when trying to set
        a property that is not defined in the class.

        Handles assignment of enumeration choices (lists/tuples) and
        converts string representations of Quantity types to Quantity objects.
        """
        if name in ("PropertiesList", "Label", "Name") or name.startswith("_"):
            super().__setattr__(name, value)
            return

        # Handle assignment of enumeration choices (list/tuple)
        prop_type = self._property_types.get(name)
        if prop_type == "App::PropertyEnumeration" and isinstance(value, (list, tuple)):
            self._property_enums[name] = list(value)
            assert len(value) > 0, f"Enum property '{name}' must have at least one entry"
            self._properties.setdefault(name, value[0])
            return

        # Attempt to convert string values to Quantity if the property type is Quantity
        elif prop_type in [
            "App::PropertyQuantity",
            "App::PropertyLength",
            "App::PropertyArea",
            "App::PropertyVolume",
            "App::PropertyAngle",
        ]:
            value = FreeCAD.Units.Quantity(value)

        # Store the (potentially converted) value
        self._properties[name] = value
        Path.Log.debug(
            f"DetachedDocumentObject: Set property '{name}' to "
            f"value {value} (type: {type(value)})"
        )

    def __getattr__(self, name: str) -> Any:
        """Intercept attribute access."""
        if name in self._properties:
            return self._properties[name]
        # Default behaviour: raise AttributeError
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def setEditorMode(self, name: str, mode: int) -> None:
        """Stores editor mode settings in detached state."""
        self._editor_modes[name] = mode

    def getEditorMode(self, name: str) -> int:
        """Stores editor mode settings in detached state."""
        return self._editor_modes.get(name, 0) or 0

    def getGroupOfProperty(self, name: str) -> Optional[str]:
        """Returns the stored group for a property in detached state."""
        return self._property_groups.get(name)

    def getTypeIdOfProperty(self, name: str) -> Optional[str]:
        """Returns the stored type string for a property in detached state."""
        return self._property_types.get(name)

    def getEnumerationsOfProperty(self, name: str) -> List[str]:
        """Returns the stored enumeration list for a property."""
        return self._property_enums.get(name, [])

    @property
    def ExpressionEngine(self) -> List[Any]:
        """Mimics the ExpressionEngine attribute of a real DocumentObject."""
        return []  # Return an empty list to satisfy iteration

    def copy_to(self, obj: FreeCAD.DocumentObject) -> None:
        """
        Copies properties from this detached object to a real DocumentObject.
        """
        for prop_name in self.PropertiesList:
            if not hasattr(self, prop_name):
                continue

            prop_value = self.getPropertyByName(prop_name)
            prop_type = self._property_types.get(prop_name)
            prop_group = self._property_groups.get(prop_name)
            prop_doc = self._property_docs.get(prop_name, "")
            prop_editor_mode = self._editor_modes.get(prop_name)

            # If the property doesn't exist in the target object, add it
            if not hasattr(obj, prop_name):
                # For enums, addProperty expects "App::PropertyEnumeration"
                # The list of choices is set afterwards.
                obj.addProperty(prop_type, prop_name, prop_group, prop_doc)

            # If it's an enumeration, set its list of choices first
            if prop_type == "App::PropertyEnumeration":
                enum_choices = self._property_enums.get(prop_name)
                assert enum_choices is not None
                setattr(obj, prop_name, enum_choices)

            # Set the property value and editor mode
            try:
                if prop_type == "App::PropertyEnumeration":
                    first_choice = self._property_enums[prop_name][0]
                    setattr(obj, prop_name, first_choice)
                setattr(obj, prop_name, prop_value)

            except Exception as e:
                Path.Log.error(
                    f"Error setting property {prop_name} to {prop_value} "
                    f"(type: {type(prop_value)}, expected type: {prop_type}): {e}"
                )
                raise

            if prop_editor_mode is not None:
                obj.setEditorMode(prop_name, prop_editor_mode)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods<at>gmail.com>            *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
This is a preprocessor file for the Path workbench. Its aim is to
parse the contents of a given OpenSBP file, and transform it to make it
suitable for use in a Path object. This preprocessor, once placed in the
appropriate PathScripts folder, can be used directly from inside FreeCAD,
via the GUI importer or via python scripts with:

import opensbp_pre
opensbp_pre.insert("/path/to/myfile.ngc","DocumentName")


DONE
Correctly imports single axis and multi axis moves.
Stores Jog and Feed speeds
Appends Multiaxis Feed speed to G1 moves
Jog rates don't append to G0 moves
Make single axis feed rates work
Imports CG (non-diameter) arcs.
Handles CW and CCW spindle speeds
if operations are preceded by a comment ('New Path ...)  They are split into multiple paths

TODO
Many other OpenSBP commands not handled

"""

import FreeCAD
import Path
import os
import Path
from builtins import open as pyopen

AXIS = (
    "X",
    "Y",
    "Z",
    "A",
    "B",
)  # OpenSBP always puts multiaxis move parameters in this order
SPEEDS = "XY", "Z", "A", "B"


def open(filename):
    "called when freecad opens a file."
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    insert(filename, doc.Name)


def insert(filename, docname):
    """called when freecad imports a file
    This insert expects parse to return a list of strings
    each string will become a separate path"""
    gfile = pyopen(filename)
    gcode = gfile.read()
    gfile.close()
    gcode = parse(gcode)
    doc = FreeCAD.getDocument(docname)
    for subpath in gcode:
        obj = doc.addObject("Path::Feature", "Path")
        path = Path.Path(subpath)
        obj.Path = path


def parse(inputstring):
    "parse(inputstring): returns a list of parsed output string"
    print("preprocessing...")
    # split the input by line
    lines = inputstring.split("\n")
    return_output = []
    output = ""
    last = {"X": None, "Y": None, "Z": None, "A": None, "B": None}
    lastrapidspeed = {
        "XY": "50",
        "Z": "50",
        "A": "50",
        "B": "50",
    }  # set default rapid speeds
    lastfeedspeed = {
        "XY": "50",
        "Z": "50",
        "A": "50",
        "B": "50",
    }  # set default feed speed
    movecommand = ["G1", "G0", "G02", "G03"]

    for line in lines:
        # remove any leftover trailing and preceding spaces
        line = line.strip()
        if not line:
            # discard empty lines
            continue
        if line[0] in ["'", "&"]:
            # discard comment and other non strictly gcode lines
            if line[0:9] == "'New Path":
                # starting new path
                if any(
                    x in output for x in movecommand
                ):  # make sure the path has at least one move command.
                    return_output.append(output)
                    output = ""
            continue

        words = [a.strip() for a in line.split(",")]
        words[0] = words[0].upper()
        if words[0] in [
            "J2",
            "J3",
            "J4",
            "J5",
            "M2",
            "M3",
            "M4",
            "M5",
        ]:  # multi-axis jogs and moves
            if words[0][0] == "J":  # jog move
                s = "G0 "
            else:  # feed move
                s = "G1 "
            speed = lastfeedspeed["XY"]

            for i in range(1, len(words)):
                if words[i] == "":
                    if last[AXIS[i - 1]] is None:
                        continue
                    else:
                        s += AXIS[i - 1] + last[AXIS[i - 1]]
                else:
                    s += AXIS[i - 1] + words[i]
                    last[AXIS[i - 1]] = words[i]
            output += s + " F" + speed + "\n"

        if words[0] in [
            "JA",
            "JB",
            "JX",
            "JY",
            "JZ",
            "MA",
            "MB",
            "MX",
            "MY",
            "MZ",
        ]:  # single axis jogs and moves
            if words[0][0] == "J":  # jog move
                s = "G0 "
                if words[0][1] in ["X", "Y"]:
                    speed = lastrapidspeed["XY"]
                else:
                    speed = lastrapidspeed[words[0][1]]

            else:  # feed move
                s = "G1 "
                if words[0][1] in ["X", "Y"]:
                    speed = lastfeedspeed["XY"]
                else:
                    speed = lastfeedspeed[words[0][1]]

            last[words[0][1]] = words[1]
            output += s
            for key, val in last.items():
                if val is not None:
                    output += key + str(val) + " F" + speed + "\n"

        if words[0] in ["JS"]:  # set jog speed
            for i in range(1, len(words)):
                if words[i] == "":
                    continue
                else:
                    lastrapidspeed[SPEEDS[i - 1]] = words[i]

        if words[0] in ["MD"]:  # move distance with distance and angle.
            # unsupported at this time
            continue
        if words[0] in ["MH"]:  # move home
            # unsupported at this time
            continue
        if words[0] in ["MS"]:  # set move speed
            for i in range(1, len(words)):
                if words[i] == "":
                    continue
                else:
                    lastfeedspeed[SPEEDS[i - 1]] = words[i]
        if words[0] in ["MO"]:  # motors off
            # unsupported at this time
            continue

        if words[0] in ["TR"]:  # Setting spindle speed
            if float(words[1]) < 0:
                s = "M4 S"
            else:
                s = "M3 S"
            s += str(abs(float(words[1])))
            output += s + "\n"

        if words[0] in ["CG"]:  # Gcode circle/arc
            if words[1] != "":  # diameter mode
                print("diameter mode not supported")
                continue

            else:
                if words[7] == "1":  # CW
                    s = "G2"
                else:  # CCW
                    s = "G3"

                s += (
                    " X"
                    + words[2]
                    + " Y"
                    + words[3]
                    + " I"
                    + words[4]
                    + " J"
                    + words[5]
                    + " F"
                    + str(lastfeedspeed["XY"])
                )
                output += s + "\n"

                last["X"] = words[2]
                last["Y"] = words[3]

    # Make sure all appended paths have at least one move command.
    if any(x in output for x in movecommand):
        return_output.append(output)
        print("done preprocessing.")

    return return_output


print(__name__ + " gcode preprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint electric charge density task panel"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package task_constraint_electricchargedensity
#  \ingroup FEM
#  \brief task panel for constraint electric charge density object

from PySide import QtCore

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets

from femtools import membertools
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):

    def __init__(self, obj):
        super().__init__(obj)

        self.parameter_widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/ElectricChargeDensity.ui"
        )

        self.init_parameter_widget()

        QtCore.QObject.connect(
            self.parameter_widget.qsb_source_charge_density,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.source_charge_density_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_interface_charge_density,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.interface_charge_density_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_total_charge,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.total_charge_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.cb_mode,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.mode_changed,
        )

        # geometry selection widget
        # start with Solid in list!
        self.selection_widget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face", "Edge"], False, False
        )

        # form made from param and selection widget
        self.form = [self.parameter_widget, self.selection_widget]

        analysis = obj.getParentGroup()
        self._mesh = None
        self._part = None
        if analysis is not None:
            self._mesh = membertools.get_single_member(analysis, "Fem::FemMeshObject")
        if self._mesh is not None:
            self._part = self._mesh.Shape
        self._partVisible = None
        self._meshVisible = None

    def open(self):
        if self._mesh is not None and self._part is not None:
            self._meshVisible = self._mesh.ViewObject.isVisible()
            self._partVisible = self._part.ViewObject.isVisible()
            self._mesh.ViewObject.hide()
            self._part.ViewObject.show()

    def reject(self):
        self.restore_visibility()
        self.selection_widget.finish_selection()
        return super().reject()

    def accept(self):
        self.obj.References = self.selection_widget.references
        self.obj.SourceChargeDensity = self.source_charge_density
        self.obj.InterfaceChargeDensity = self.interface_charge_density
        self.obj.TotalCharge = self.total_charge
        self.obj.Mode = self.mode

        self.selection_widget.finish_selection()
        self.restore_visibility()
        return super().accept()

    def restore_visibility(self):
        if self._mesh is not None and self._part is not None:
            if self._meshVisible:
                self._mesh.ViewObject.show()
            else:
                self._mesh.ViewObject.hide()
            if self._partVisible:
                self._part.ViewObject.show()
            else:
                self._part.ViewObject.hide()

    def init_parameter_widget(self):
        self.source_charge_density = self.obj.SourceChargeDensity
        self.interface_charge_density = self.obj.InterfaceChargeDensity
        self.total_charge = self.obj.TotalCharge
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_source_charge_density).bind(
            self.obj, "SourceChargeDensity"
        )
        self.parameter_widget.qsb_source_charge_density.setProperty(
            "value", self.source_charge_density
        )

        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_interface_charge_density).bind(
            self.obj, "InterfaceChargeDensity"
        )
        self.parameter_widget.qsb_interface_charge_density.setProperty(
            "value", self.interface_charge_density
        )

        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_total_charge).bind(
            self.obj, "TotalCharge"
        )
        self.parameter_widget.qsb_total_charge.setProperty("value", self.total_charge)

        self.mode = self.obj.Mode
        self.mode_enum = self.obj.getEnumerationsOfProperty("Mode")
        self.parameter_widget.cb_mode.addItems(self.mode_enum)
        index = self.mode_enum.index(self.mode)
        self.parameter_widget.cb_mode.setCurrentIndex(index)
        self.mode_changed(index)

    def source_charge_density_changed(self, base_quantity_value):
        self.source_charge_density = base_quantity_value

    def interface_charge_density_changed(self, base_quantity_value):
        self.interface_charge_density = base_quantity_value

    def total_charge_changed(self, base_quantity_value):
        self.total_charge = base_quantity_value

    def mode_changed(self, index):
        self.mode = self.mode_enum[index]
        if self.mode in ["Total Interface", "Total Source"]:
            index = 2
        self.parameter_widget.sw_mode.setCurrentIndex(index)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ****************************************************************************
# *  Copyright (c) 2018 Maurice <easyw@katamail.com>                         *
# *                                                                          *
# *  StepZ Import Export compressed STEP files for FreeCAD                   *
# *  License: LGPLv2+                                                        *
# *                                                                          *
# ****************************************************************************

# workaround for unicode in gzipping filename
# OCC7 doesn't support non-ASCII characters at the moment
# https://forum.freecad.org/viewtopic.php?t=20815


import FreeCAD
import FreeCADGui
import shutil
import os
import re
import ImportGui
import PySide
from PySide import QtCore
from PySide import QtGui
import tempfile

___stpZversion___ = "1.4.0"
# support both gz and zipfile archives
# Catia seems to use gz, Inventor zipfile
# improved import, open and export


import gzip as gz
import builtins
import importlib


import zipfile as zf

# import stepZ; import importlib; importlib.reload(stepZ); stepZ.open(u"C:/Temp/brick.stpz")


def mkz_string(input):
    if isinstance(input, str):
        return input
    else:
        input = input.encode("utf-8")
        return input


####
def mkz_unicode(input):
    if isinstance(input, str):
        return input
    else:
        input = input.decode("utf-8")
        return input


####
def sayz(msg):
    FreeCAD.Console.PrintMessage(msg)
    FreeCAD.Console.PrintMessage("\n")


####
def sayzw(msg):
    FreeCAD.Console.PrintWarning(msg)
    FreeCAD.Console.PrintWarning("\n")


####
def sayzerr(msg):
    FreeCAD.Console.PrintError(msg)
    FreeCAD.Console.PrintWarning("\n")


####


def import_stpz(fn, fc, doc):

    # sayz(fn)
    ext = os.path.splitext(os.path.basename(fn))[1]
    fname = os.path.splitext(os.path.basename(fn))[0]
    basepath = os.path.split(fn)[0]
    filepath = os.path.join(basepath, fname + ".stp")

    tempdir = tempfile.gettempdir()  # get the current temporary directory
    tempfilepath = os.path.join(tempdir, fname + ".stp")

    with builtins.open(tempfilepath, "wb") as f:  # py3
        f.write(fc)
    # ImportGui.insert(filepath)
    if doc is None:
        ImportGui.open(tempfilepath)
    else:
        ImportGui.open(tempfilepath, doc.Name)
    FreeCADGui.SendMsgToActiveView("ViewFit")
    try:
        os.remove(tempfilepath)
    except OSError:
        sayzerr("error on removing " + tempfilepath + " file")


###


def open(filename, doc=None):

    if zf.is_zipfile(filename):
        with zf.ZipFile(filename, "r") as fz:
            file_names = fz.namelist()
            for fn in file_names:
                sayz(fn)
                with fz.open(fn) as zfile:
                    file_content = zfile.read()
                    import_stpz(filename, file_content, doc)
    else:
        with gz.open(filename, "rb") as f:
            fnm = os.path.splitext(os.path.basename(filename))[0]
            sayz(fnm)
            file_content = f.read()
            import_stpz(filename, file_content, doc)


####


def insert(filename, doc):

    doc = FreeCAD.ActiveDocument
    open(filename, doc)


####


def export(objs, filename):
    """exporting to file folder"""

    # sayz(filename)
    sayz("stpZ version " + ___stpZversion___)
    ext = os.path.splitext(os.path.basename(filename))[1]
    fname = os.path.splitext(os.path.basename(filename))[0]
    basepath = os.path.split(filename)[0]
    tempdir = tempfile.gettempdir()  # get the current temporary directory

    filepath = os.path.join(basepath, fname) + ".stp"
    filepath_base = os.path.join(basepath, fname)

    namefpath = os.path.join(basepath, fname)

    outfpath = os.path.join(basepath, fname) + ".stpZ"
    outfpathT = os.path.join(tempdir, fname) + ".stpZ"
    outfpath_stp = os.path.join(basepath, fname) + ".stp"
    outfpathT_stp = os.path.join(tempdir, fname) + ".stp"

    outfpath_base = basepath
    # outfpath_str = mkz_string(os.path.join(basepath,fname))
    outfpath_str = os.path.join(basepath, fname) + ".stp"
    outfpathT_str = os.path.join(tempdir, fname) + ".stp"

    if os.path.exists(outfpathT_stp):
        os.remove(outfpathT_stp)
        sayzw("Old temp file with the same name removed '" + outfpathT_stp + "'")
    ImportGui.export(objs, outfpathT_stp)
    with builtins.open(outfpathT_stp, "rb") as f_in:
        file_content = f_in.read()
        new_f_content = file_content
        f_in.close()
    with gz.open(outfpathT_str, "wb") as f_out:
        f_out.write(new_f_content)
        f_out.close()
    if os.path.exists(outfpath):
        shutil.move(outfpathT_str, outfpath)
        # os.remove(outfpathT_stp)
    else:
        shutil.move(outfpathT_str, outfpath)
        # os.remove(outfpathT_stp)


####



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
import asyncio
import hashlib
from typing import Any, Type, Optional, List, Mapping
from Path.Tool.assets.cache import AssetCache, CacheKey
from Path.Tool.assets import (
    AssetManager,
    Asset,
    AssetUri,
    AssetSerializer,
    DummyAssetSerializer,
    MemoryStore,
)


class MockAsset(Asset):
    asset_type: str = "mock_asset"
    _build_counter = 0

    def __init__(
        self,
        asset_id: str,
        raw_data: bytes,
        dependencies: Optional[Mapping[AssetUri, Any]] = None,
    ):
        super().__init__()  # Initialize Asset ABC
        self._asset_id = asset_id  # Store id internally
        self.raw_data_content = raw_data
        self.resolved_dependencies = dependencies or {}
        MockAsset._build_counter += 1
        self.build_id = MockAsset._build_counter

    def get_id(self) -> str:  # Implement abstract method
        return self._asset_id

    # get_uri() is inherited from Asset and uses self.asset_type and self.get_id()

    @classmethod
    def extract_dependencies(cls, data: bytes, serializer: AssetSerializer) -> List[AssetUri]:
        """Extracts URIs of dependencies from serialized data."""
        # This mock implementation handles the simple "dep:" format
        data_str = data.decode()
        if data_str.startswith("dep:"):
            try:
                # Get content after the first "dep:"
                dep_content = data_str.split(":", 1)[1]
            except IndexError:
                # This case should ideally not be reached if startswith("dep:") is true
                # and there's content after "dep:", but good for robustness.
                return []

            dep_uri_strings = dep_content.split(",")
            uris = []
            for uri_string in dep_uri_strings:
                uri_string = uri_string.strip()  # Remove leading/trailing whitespace
                if not uri_string:
                    continue
                try:
                    uris.append(AssetUri(uri_string))
                except ValueError:
                    # This print will now show the full problematic uri_string
                    print(f"Warning: Could not parse mock dependency URI: '{uri_string}'")
            return uris
        return []

    @classmethod
    def from_bytes(
        cls: Type["MockAsset"],
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Any]],
        serializer: AssetSerializer,
    ) -> "MockAsset":
        return cls(asset_id=id, raw_data=data, dependencies=dependencies)

    def to_bytes(self, serializer: AssetSerializer) -> bytes:  # Implement abstract method
        return self.raw_data_content


class MockAssetB(Asset):  # New mock asset class for type 'mock_asset_b'
    asset_type: str = "mock_asset_b"
    _build_counter = 0  # Separate counter if needed, or share MockAsset's

    def __init__(
        self,
        asset_id: str,
        raw_data: bytes,
        dependencies: Optional[Mapping[AssetUri, Any]] = None,
    ):
        super().__init__()
        self._asset_id = asset_id
        self.raw_data_content = raw_data
        self.resolved_dependencies = dependencies or {}
        MockAssetB._build_counter += 1
        self.build_id = MockAssetB._build_counter

    def get_id(self) -> str:
        return self._asset_id

    @classmethod
    def extract_dependencies(cls, data: bytes, serializer: AssetSerializer) -> List[AssetUri]:
        # Keep simple, or adapt if MockAssetB has different dep logic
        return []

    @classmethod
    def from_bytes(
        cls: Type["MockAssetB"],
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Any]],
        serializer: AssetSerializer,
    ) -> "MockAssetB":
        return cls(asset_id=id, raw_data=data, dependencies=dependencies)

    def to_bytes(self, serializer: AssetSerializer) -> bytes:
        return self.raw_data_content


def _get_raw_data_hash(raw_data: bytes) -> int:
    return int(hashlib.sha256(raw_data).hexdigest(), 16)


class TestPathToolAssetCache(unittest.TestCase):
    def setUp(self):
        self.cache = AssetCache(max_size_bytes=1000)

    def test_put_and_get_simple(self):
        key = CacheKey("store1", "mock_asset://id1", _get_raw_data_hash(b"data1"), ("dep1",))
        asset_obj = MockAsset(asset_id="id1", raw_data=b"data1")
        self.cache.put(key, asset_obj, len(b"data1"), {"dep_uri_str"})

        retrieved = self.cache.get(key)
        self.assertIsNotNone(retrieved)
        # Assuming retrieved is MockAsset, it will have get_id()
        self.assertEqual(retrieved.get_id(), "id1")

    def test_get_miss(self):
        key = CacheKey("store1", "mock_asset://id1", _get_raw_data_hash(b"data1"), tuple())
        self.assertIsNone(self.cache.get(key))

    def test_lru_eviction(self):
        asset_data_size = 300
        asset1_data = b"a" * asset_data_size
        asset2_data = b"b" * asset_data_size
        asset3_data = b"c" * asset_data_size
        asset4_data = b"d" * asset_data_size

        key1 = CacheKey("s", "mock_asset://id1", _get_raw_data_hash(asset1_data), tuple())
        key2 = CacheKey("s", "mock_asset://id2", _get_raw_data_hash(asset2_data), tuple())
        key3 = CacheKey("s", "mock_asset://id3", _get_raw_data_hash(asset3_data), tuple())
        key4 = CacheKey("s", "mock_asset://id4", _get_raw_data_hash(asset4_data), tuple())

        self.cache.put(key1, MockAsset("id1", asset1_data), asset_data_size, set())
        self.cache.put(key2, MockAsset("id2", asset2_data), asset_data_size, set())
        self.cache.put(key3, MockAsset("id3", asset3_data), asset_data_size, set())

        self.assertEqual(self.cache.current_size_bytes, 3 * asset_data_size)
        self.assertIsNotNone(self.cache.get(key1))  # Access key1 to make it MRU

        # Adding key4 should evict key2 (oldest after key1 accessed)
        self.cache.put(key4, MockAsset("id4", asset4_data), asset_data_size, set())
        self.assertEqual(self.cache.current_size_bytes, 3 * asset_data_size)
        self.assertIsNotNone(self.cache.get(key1))
        self.assertIsNone(self.cache.get(key2))  # Evicted
        self.assertIsNotNone(self.cache.get(key3))
        self.assertIsNotNone(self.cache.get(key4))

    def test_invalidate_direct(self):
        key = CacheKey("s", "mock_asset://id1", _get_raw_data_hash(b"data"), tuple())
        self.cache.put(key, MockAsset("id1", b"data"), 4, set())
        retrieved = self.cache.get(key)  # Ensure it's there
        self.assertIsNotNone(retrieved)

        self.cache.invalidate_for_uri("mock_asset://id1")
        self.assertIsNone(self.cache.get(key))
        self.assertEqual(self.cache.current_size_bytes, 0)

    def test_invalidate_recursive(self):
        data_a = b"data_a_dep:mock_asset_b://idB"  # A depends on B
        data_b = b"data_b"
        data_c = b"data_c_dep:mock_asset_a://idA"  # C depends on A

        uri_a_str = "mock_asset_a://idA"
        uri_b_str = "mock_asset_b://idB"
        uri_c_str = "mock_asset_c://idC"

        key_b = CacheKey("s", uri_b_str, _get_raw_data_hash(data_b), tuple())
        key_a = CacheKey("s", uri_a_str, _get_raw_data_hash(data_a), (uri_b_str,))
        key_c = CacheKey("s", uri_c_str, _get_raw_data_hash(data_c), (uri_a_str,))

        self.cache.put(key_b, MockAsset("idB", data_b), len(data_b), set())
        self.cache.put(key_a, MockAsset("idA", data_a), len(data_a), {uri_b_str})
        self.cache.put(key_c, MockAsset("idC", data_c), len(data_c), {uri_a_str})

        self.assertIsNotNone(self.cache.get(key_a))
        self.assertIsNotNone(self.cache.get(key_b))
        self.assertIsNotNone(self.cache.get(key_c))

        self.cache.invalidate_for_uri(uri_b_str)  # Invalidate B

        self.assertIsNone(self.cache.get(key_a), "Asset A should be invalidated")
        self.assertIsNone(self.cache.get(key_b), "Asset B should be invalidated")
        self.assertIsNone(self.cache.get(key_c), "Asset C should be invalidated")
        self.assertEqual(self.cache.current_size_bytes, 0)

    def test_clear_cache(self):
        key = CacheKey("s", "mock_asset://id1", _get_raw_data_hash(b"data"), tuple())
        self.cache.put(key, MockAsset("id1", b"data"), 4, set())
        self.assertNotEqual(self.cache.current_size_bytes, 0)

        self.cache.clear()
        self.assertIsNone(self.cache.get(key))
        self.assertEqual(self.cache.current_size_bytes, 0)
        self.assertEqual(len(self.cache._cache_dependencies_map), 0)
        self.assertEqual(len(self.cache._cache_dependents_map), 0)


class TestPathToolAssetCacheIntegration(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.manager = AssetManager(cache_max_size_bytes=10 * 1024)  # 10KB cache
        self.store_name = "test_store"
        self.store = MemoryStore(name=self.store_name)
        self.manager.register_store(self.store, cacheable=True)
        self.manager.register_asset(MockAsset, DummyAssetSerializer)
        self.manager.register_asset(MockAssetB, DummyAssetSerializer)  # Register the new mock type
        MockAsset._build_counter = 0
        MockAssetB._build_counter = 0

    def tearDown(self):
        self.loop.close()

    def _run_async(self, coro):
        return self.loop.run_until_complete(coro)

    def test_get_caches_asset(self):
        uri_str = "mock_asset://asset1"
        raw_data = b"asset1_data"
        self._run_async(self.store.create("mock_asset", "asset1", raw_data))

        # First get - should build and cache
        asset1 = self.manager.get(uri_str, store=self.store_name)
        self.assertIsInstance(asset1, MockAsset)
        self.assertEqual(asset1.get_id(), "asset1")
        self.assertEqual(MockAsset._build_counter, 1)  # Built once

        # Second get - should hit cache
        asset2 = self.manager.get(uri_str, store=self.store_name)
        self.assertIsInstance(asset2, MockAsset)
        self.assertEqual(asset2.get_id(), "asset1")
        self.assertEqual(MockAsset._build_counter, 1)  # Still 1, not rebuilt
        self.assertIs(asset1, asset2)  # Should be the same instance from cache

    def test_get_respects_depth_in_cache_key(self):
        uri_str = "mock_asset://asset_depth"
        # A depends on B (mock_asset_b://dep_b)
        raw_data_a = b"dep:mock_asset_b://dep_b"
        raw_data_b = b"dep_b_data"

        self._run_async(self.store.create("mock_asset", "asset_depth", raw_data_a))
        self._run_async(self.store.create("mock_asset_b", "dep_b", raw_data_b))

        # Get with depth=0 (shallow)
        asset_shallow = self.manager.get(uri_str, store=self.store_name, depth=0)
        self.assertIsInstance(asset_shallow, MockAsset)
        self.assertEqual(len(asset_shallow.resolved_dependencies), 0)
        self.assertEqual(MockAsset._build_counter, 1)  # asset_depth built

        # Get with depth=None (full)
        asset_full = self.manager.get(uri_str, store=self.store_name, depth=None)
        self.assertIsInstance(asset_full, MockAsset)
        self.assertEqual(len(asset_full.resolved_dependencies), 1)
        # asset_depth (MockAsset) built twice (once shallow, once full)
        self.assertEqual(MockAsset._build_counter, 2)
        # dep_b (MockAssetB) built once as a dependency of the full asset_depth
        self.assertEqual(MockAssetB._build_counter, 1)

        # Get shallow again - should hit shallow cache
        asset_shallow_2 = self.manager.get(uri_str, store=self.store_name, depth=0)
        self.assertIs(asset_shallow, asset_shallow_2)
        self.assertEqual(MockAsset._build_counter, 2)  # No new MockAsset builds
        self.assertEqual(MockAssetB._build_counter, 1)  # No new MockAssetB builds

        # Get full again - should hit full cache
        asset_full_2 = self.manager.get(uri_str, store=self.store_name, depth=None)
        self.assertIs(asset_full, asset_full_2)
        self.assertEqual(MockAsset._build_counter, 2)  # No new MockAsset builds
        self.assertEqual(MockAssetB._build_counter, 1)  # No new MockAssetB builds

    def test_update_invalidates_cache(self):
        uri_str = "mock_asset://asset_upd"
        raw_data_v1 = b"version1"
        raw_data_v2 = b"version2"
        asset_uri = AssetUri(uri_str)  # Use real AssetUri

        self._run_async(self.store.create(asset_uri.asset_type, asset_uri.asset_id, raw_data_v1))

        asset_v1 = self.manager.get(asset_uri, store=self.store_name)
        self.assertEqual(asset_v1.raw_data_content, raw_data_v1)
        self.assertEqual(MockAsset._build_counter, 1)

        # Update the asset in the store (MemoryStore creates new version)
        # For this test, let's simulate an update by re-adding with add_raw
        # which should trigger invalidation.
        # Note: MemoryStore's update creates a new version, so get() would get latest.
        # To test invalidation of the *exact* cached object, we'd need to ensure
        # the cache key changes (e.g. raw_data_hash).
        # Let's use add_raw which calls invalidate.
        self.manager.add_raw(
            asset_uri.asset_type, asset_uri.asset_id, raw_data_v2, store=self.store_name
        )

        # Get again - should rebuild because v1 was invalidated
        # And MemoryStore's get() for a URI without version gets latest.
        # The add_raw invalidates based on URI (type+id), so all versions of it.
        asset_v2 = self.manager.get(asset_uri, store=self.store_name)
        self.assertEqual(asset_v2.raw_data_content, raw_data_v2)
        self.assertEqual(MockAsset._build_counter, 2)  # Rebuilt

    def test_delete_invalidates_cache(self):
        uri_str = "mock_asset://asset_del"
        raw_data = b"delete_me"
        asset_uri = AssetUri(uri_str)
        self._run_async(self.store.create(asset_uri.asset_type, asset_uri.asset_id, raw_data))

        asset1 = self.manager.get(asset_uri, store=self.store_name)
        self.assertIsNotNone(asset1)
        self.assertEqual(MockAsset._build_counter, 1)

        self.manager.delete(asset_uri, store=self.store_name)

        with self.assertRaises(FileNotFoundError):
            self.manager.get(asset_uri, store=self.store_name)
        # Check build counter didn't increase due to trying to get deleted asset
        self.assertEqual(MockAsset._build_counter, 1)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2020 Schildkroet                                        *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""ToolBit Library Dock Widget."""
import FreeCAD
import FreeCADGui
import Path
import Path.Tool.Gui.Controller as PathToolControllerGui
import PathScripts.PathUtilsGui as PathUtilsGui
from PySide import QtGui, QtCore, QtWidgets
from functools import partial
from typing import List, Tuple
from ...camassets import cam_assets, ensure_assets_initialized
from ...toolbit import ToolBit
from .editor import LibraryEditor
from .browser import LibraryBrowserWithCombo

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


class ToolBitLibraryDock(object):
    """Controller for displaying a library and creating ToolControllers"""

    def __init__(self, defaultJob=None, autoClose=False):
        ensure_assets_initialized(cam_assets)
        # Create the main form widget directly
        self.defaultJob = defaultJob
        self.autoClose = autoClose
        self.form = QtWidgets.QDialog()
        self.form.setObjectName("ToolSelector")
        self.form.setWindowTitle(translate("CAM_ToolBit", "Toolbit Selector"))
        self.form.setMinimumSize(600, 400)
        self.form.resize(800, 600)
        self.form.adjustSize()
        self.form_layout = QtGui.QVBoxLayout(self.form)
        self.form_layout.setContentsMargins(4, 4, 4, 4)
        self.form_layout.setSpacing(4)

        # Create the browser widget
        self.browser_widget = LibraryBrowserWithCombo(asset_manager=cam_assets)

        self._setup_ui()

    def _setup_ui(self):
        """Setup the form and load the tooltable data"""
        Path.Log.track()

        # Create a main widget and layout for the dock
        main_widget = QtGui.QWidget()
        main_layout = QtGui.QVBoxLayout(main_widget)
        main_layout.setContentsMargins(4, 4, 4, 4)
        main_layout.setSpacing(4)

        main_layout.addWidget(self.browser_widget)

        # Create buttons
        self.libraryEditorOpenButton = QtGui.QPushButton(
            translate("CAM_ToolBit", "Open Library Editor")
        )
        self.addToolControllerButton = QtGui.QPushButton(translate("CAM_ToolBit", "Add to Job"))
        self.closeButton = QtGui.QPushButton(translate("CAM_ToolBit", "Close"))

        button_width = 120
        self.libraryEditorOpenButton.setMinimumWidth(button_width)
        self.addToolControllerButton.setMinimumWidth(button_width)
        self.closeButton.setMinimumWidth(button_width)

        # Add buttons to a horizontal layout, right-align Close
        button_layout = QtGui.QHBoxLayout()
        button_layout.addWidget(self.libraryEditorOpenButton)
        button_layout.addWidget(self.addToolControllerButton)
        button_layout.addStretch(1)
        button_layout.addWidget(self.closeButton)

        # Add the button layout to the main layout
        main_layout.addLayout(button_layout)

        # Set the main widget as the dock's widget
        self.form.layout().addWidget(main_widget)

        # Connect signals from the browser widget and buttons
        self.browser_widget.toolSelected.connect(lambda x: self._update_state())
        self.browser_widget.itemDoubleClicked.connect(self._on_doubleclick)
        self.libraryEditorOpenButton.clicked.connect(self._open_editor)
        self.addToolControllerButton.clicked.connect(self._add_tool_controller_to_doc)
        self.closeButton.clicked.connect(self.form.reject)

        # Update the initial state of the UI
        self._update_state()

    def _count_jobs(self):
        if not FreeCAD.ActiveDocument:
            return 0
        return len([1 for j in FreeCAD.ActiveDocument.Objects if j.Name[:3] == "Job"]) >= 1

    def _update_state(self):
        """Enable button to add tool controller when a tool is selected"""
        selected = bool(self.browser_widget.get_selected_bit_uris())
        has_job = selected and self._count_jobs() > 0
        self.addToolControllerButton.setEnabled(selected and has_job)

    def _on_doubleclick(self, toolbit: ToolBit):
        """Opens the ToolBitEditor for the selected toolbit."""
        self._add_tool_controller_to_doc()

    def _open_editor(self):
        library = LibraryEditor(parent=FreeCADGui.getMainWindow())
        library.open()
        # After editing, we might need to refresh the libraries in the browser widget
        # Assuming _populate_libraries is the correct method to call
        self.browser_widget.refresh()

    def _add_tool_to_doc(self) -> List[Tuple[int, ToolBit]]:
        """
        Get the selected toolbit assets from the browser widget.
        """
        Path.Log.track()
        tools = []
        selected_toolbits = self.browser_widget.get_selected_bits()

        for toolbit in selected_toolbits:
            # Need to get the tool number for this toolbit from the currently
            # selected library in the browser widget.
            toolNr = self.browser_widget.get_tool_no_from_current_library(toolbit)
            if toolNr is not None:
                toolbit.attach_to_doc(FreeCAD.ActiveDocument)
                tools.append((toolNr, toolbit))
            else:
                Path.Log.warning(
                    f"Could not get tool number for toolbit {toolbit.get_uri()} in selected library."
                )

        return tools

    def _add_tool_controller_to_doc(self):
        """
        if no jobs, don't do anything, otherwise all TCs for all
        selected toolbit assets
        """
        Path.Log.track()
        jobs = PathUtilsGui.PathUtils.GetJobs()
        if len(jobs) == 0:
            QtGui.QMessageBox.information(
                self.form,
                translate("CAM_ToolBit", "No Job Found"),
                translate("CAM_ToolBit", "Please create a Job first."),
            )
            return
        elif self.defaultJob or len(jobs) == 1:
            job = self.defaultJob or jobs[0]
        else:
            userinput = PathUtilsGui.PathUtilsUserInput()
            job = userinput.chooseJob(jobs)
            self.defaultJob = job

        if job is None:  # user may have canceled
            return

        # Get the selected toolbit assets
        selected_tools = self._add_tool_to_doc()

        for toolNr, toolbit in selected_tools:
            tc = PathToolControllerGui.Create(f"TC: {toolbit.label}", toolbit.obj, toolNr)
            job.Proxy.addToolController(tc)
            FreeCAD.ActiveDocument.recompute()

        if self.autoClose:
            self.form.accept()

    def open(self, path=None):
        """load library stored in path and bring up ui"""
        self.form.exec_()



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Wrapper around git executable to simplify calling git commands from Python."""

# pylint: disable=too-few-public-methods

import os
import platform
import shutil
import subprocess
from typing import List, Dict, Optional
import time

import addonmanager_utilities as utils
import addonmanager_freecad_interface as fci

translate = fci.translate


class NoGitFound(RuntimeError):
    """Could not locate the git executable on this system."""


class GitFailed(RuntimeError):
    """The call to git returned an error of some kind"""


def _ref_format_string() -> str:
    return (
        "--format=%(refname:lstrip=2)\t%(upstream:lstrip=2)\t%(authordate:rfc)\t%("
        "authorname)\t%(subject)"
    )


def _parse_ref_table(text: str):
    rows = text.splitlines()
    result = []
    for row in rows:
        columns = row.split("\t")
        result.append(
            {
                "ref_name": columns[0],
                "upstream": columns[1],
                "date": columns[2],
                "author": columns[3],
                "subject": columns[4],
            }
        )
    return result


class GitManager:
    """A class to manage access to git: mostly just provides a simple wrapper around
    the basic command-line calls. Provides optional asynchronous access to clone and
    update."""

    def __init__(self):
        self.git_exe = None
        self._find_git()
        if not self.git_exe:
            raise NoGitFound()

    def clone(self, remote, local_path, args: List[str] = None):
        """Clones the remote to the local path"""
        final_args = ["clone", "--recurse-submodules"]
        if args:
            final_args.extend(args)
        final_args.extend([remote, local_path])
        self._synchronous_call_git(final_args)

    def async_clone(self, remote, local_path, progress_monitor, args: List[str] = None):
        """Clones the remote to the local path, sending periodic progress updates
        to the passed progress_monitor. Returns a handle that can be used to
        cancel the job."""

    def checkout(self, local_path, spec, args: List[str] = None):
        """Checks out a specific git revision, tag, or branch. Any valid argument to
        git checkout can be submitted."""
        old_dir = os.getcwd()
        os.chdir(local_path)
        final_args = ["checkout"]
        if args:
            final_args.extend(args)
        final_args.append(spec)
        self._synchronous_call_git(final_args)
        os.chdir(old_dir)

    def dirty(self, local_path: str) -> bool:
        """Check for local changes"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        result = False
        final_args = ["diff-index", "HEAD"]
        try:
            stdout = self._synchronous_call_git(final_args)
            if stdout:
                result = True
        except GitFailed:
            result = False
        os.chdir(old_dir)
        return result

    def detached_head(self, local_path: str) -> bool:
        """Check for detached head state"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        result = False
        final_args = ["rev-parse", "--abbrev-ref", "--symbolic-full-name", "HEAD"]
        try:
            stdout = self._synchronous_call_git(final_args)
            if stdout == "HEAD":
                result = True
        except GitFailed:
            result = False
        os.chdir(old_dir)
        return result

    def update(self, local_path):
        """Fetches and pulls the local_path from its remote"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            self._synchronous_call_git(["fetch"])
            self._synchronous_call_git(["pull"])
            self._synchronous_call_git(["submodule", "update", "--init", "--recursive"])
        except GitFailed as e:
            fci.Console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Basic Git update failed with the following message:",
                )
                + str(e)
                + "\n"
            )
            fci.Console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Backing up the original directory and re-cloning",
                )
                + "...\n"
            )
            remote = self.get_remote(local_path)
            with open(os.path.join(local_path, "ADDON_DISABLED"), "w", encoding="utf-8") as f:
                f.write(
                    "This is a backup of an addon that failed to update cleanly so "
                    "was re-cloned. It was disabled by the Addon Manager's git update "
                    "facility and can be safely deleted if the addon is working "
                    "properly."
                )
            os.chdir("..")
            os.rename(local_path, local_path + ".backup" + str(time.time()))
            self.clone(remote, local_path)
        os.chdir(old_dir)

    def status(self, local_path) -> str:
        """Gets the v1 porcelain status"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            status = self._synchronous_call_git(["status", "-sb", "--porcelain"])
        except GitFailed as e:
            os.chdir(old_dir)
            raise e

        os.chdir(old_dir)
        return status

    def reset(self, local_path, args: List[str] = None):
        """Executes the git reset command"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        final_args = ["reset"]
        if args:
            final_args.extend(args)
        try:
            self._synchronous_call_git(final_args)
        except GitFailed as e:
            os.chdir(old_dir)
            raise e
        os.chdir(old_dir)

    def async_fetch_and_update(self, local_path, progress_monitor, args=None):
        """Same as fetch_and_update, but asynchronous"""

    def update_available(self, local_path) -> bool:
        """Returns True if an update is available from the remote, or false if not"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            self._synchronous_call_git(["fetch"])
            status = self._synchronous_call_git(["status", "-sb", "--porcelain"])
        except GitFailed as e:
            os.chdir(old_dir)
            raise e
        os.chdir(old_dir)
        return "behind" in status

    def current_tag(self, local_path) -> str:
        """Get the name of the currently checked-out tag if HEAD is detached"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            tag = self._synchronous_call_git(["describe", "--tags"]).strip()
        except GitFailed as e:
            os.chdir(old_dir)
            raise e
        os.chdir(old_dir)
        return tag

    def current_branch(self, local_path) -> str:
        """Get the name of the current branch"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            # This only works with git 2.22 and later (June 2019)
            # branch = self._synchronous_call_git(["branch", "--show-current"]).strip()

            # This is more universal (albeit more opaque to the reader):
            branch = self._synchronous_call_git(["rev-parse", "--abbrev-ref", "HEAD"]).strip()
        except GitFailed as e:
            os.chdir(old_dir)
            raise e
        os.chdir(old_dir)
        return branch

    def repair(self, remote, local_path):
        """Assumes that local_path is supposed to be a local clone of the given
        remote, and ensures that it is. Note that any local changes in local_path
        will be destroyed. This is achieved by archiving the old path, cloning an
        entirely new copy, and then deleting the old directory."""

        original_cwd = os.getcwd()

        # Make sure we are not currently in that directory, otherwise on Windows the
        # "rename" will fail. To guarantee we aren't in it, change to it, then shift
        # up one.
        os.chdir(local_path)
        os.chdir("..")
        backup_path = local_path + ".backup" + str(time.time())
        os.rename(local_path, backup_path)
        try:
            self.clone(remote, local_path)
        except GitFailed as e:
            fci.Console.PrintError(
                translate("AddonsInstaller", "Failed to clone {} into {} using Git").format(
                    remote, local_path
                )
            )
            os.chdir(original_cwd)
            raise e
        os.chdir(original_cwd)
        shutil.rmtree(backup_path, ignore_errors=True)

    def get_remote(self, local_path) -> str:
        """Get the repository that this local path is set to fetch from"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            response = self._synchronous_call_git(["remote", "-v", "show"])
        except GitFailed as e:
            os.chdir(old_dir)
            raise e
        lines = response.split("\n")
        result = "(unknown remote)"
        for line in lines:
            if line.endswith("(fetch)"):
                # The line looks like:
                # origin  https://some/sort/of/path (fetch)

                segments = line.split()
                if len(segments) == 3:
                    result = segments[1]
                    break
                fci.Console.PrintWarning("Error parsing the results from git remote -v show:\n")
                fci.Console.PrintWarning(line + "\n")
        os.chdir(old_dir)
        return result

    def get_branches(self, local_path) -> List[str]:
        """Get a list of all available branches (local and remote)"""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            stdout = self._synchronous_call_git(["branch", "-a", "--format=%(refname:lstrip=2)"])
        except GitFailed as e:
            os.chdir(old_dir)
            raise e
        os.chdir(old_dir)
        branches = []
        for branch in stdout.split("\n"):
            branches.append(branch)
        return branches

    def get_branches_with_info(self, local_path) -> List[Dict[str, str]]:
        """Get a list of branches, where each entry is a dictionary with status information about
        the branch."""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            stdout = self._synchronous_call_git(["branch", "-a", _ref_format_string()])
            return _parse_ref_table(stdout)
        except GitFailed as e:
            os.chdir(old_dir)
            raise e

    def get_tags_with_info(self, local_path) -> List[Dict[str, str]]:
        """Get a list of branches, where each entry is a dictionary with status information about
        the branch."""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            stdout = self._synchronous_call_git(["tag", "-l", _ref_format_string()])
            return _parse_ref_table(stdout)
        except GitFailed as e:
            os.chdir(old_dir)
            raise e

    def get_last_committers(self, local_path, n=10):
        """Examine the last n entries of the commit history, and return a list of all
        the committers, their email addresses, and how many commits each one is
        responsible for.
        """
        old_dir = os.getcwd()
        os.chdir(local_path)
        authors = self._synchronous_call_git(["log", f"-{n}", "--format=%cN"]).split("\n")
        emails = self._synchronous_call_git(["log", f"-{n}", "--format=%cE"]).split("\n")
        os.chdir(old_dir)

        result_dict = {}
        for author, email in zip(authors, emails):
            if not author or not email:
                continue
            if author not in result_dict:
                result_dict[author] = {}
                result_dict[author]["email"] = [email]
                result_dict[author]["count"] = 1
            else:
                if email not in result_dict[author]["email"]:
                    # Same author name, new email address -- treat it as the same
                    # person with a second email, instead of as a whole new person
                    result_dict[author]["email"].append(email)
                result_dict[author]["count"] += 1
        return result_dict

    def get_last_authors(self, local_path, n=10):
        """Examine the last n entries of the commit history, and return a list of all
        the authors, their email addresses, and how many commits each one is
        responsible for.
        """
        old_dir = os.getcwd()
        os.chdir(local_path)
        authors = self._synchronous_call_git(["log", f"-{n}", "--format=%aN"])
        emails = self._synchronous_call_git(["log", f"-{n}", "--format=%aE"])
        os.chdir(old_dir)

        result_dict = {}
        for author, email in zip(authors, emails):
            if author not in result_dict:
                result_dict[author]["email"] = [email]
                result_dict[author]["count"] = 1
            else:
                if email not in result_dict[author]["email"]:
                    # Same author name, new email address -- treat it as the same
                    # person with a second email, instead of as a whole new person
                    result_dict[author]["email"].append(email)
                result_dict[author]["count"] += 1
        return result_dict

    def migrate_branch(self, local_path: str, old_branch: str, new_branch: str) -> None:
        """Rename a branch (used when the remote branch name changed). Assumes that "origin"
        exists."""
        old_dir = os.getcwd()
        os.chdir(local_path)
        try:
            self._synchronous_call_git(["branch", "-m", old_branch, new_branch])
            self._synchronous_call_git(["fetch", "origin"])
            self._synchronous_call_git(["branch", "--unset-upstream"])
            self._synchronous_call_git(["branch", f"--set-upstream-to=origin/{new_branch}"])
            self._synchronous_call_git(["pull"])
        except GitFailed as e:
            fci.Console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Git branch rename failed with the following message:",
                )
                + str(e)
                + "\n"
            )
            os.chdir(old_dir)
            raise e
        os.chdir(old_dir)

    def _find_git(self):

        fc_dir = fci.DataPaths().home_dir
        git_exe = os.path.join(fc_dir, "bin", "git")
        if "Windows" in platform.system():
            git_exe += ".exe"

        if platform.system() == "Darwin" and not self._git_is_real():
            return

        if not git_exe or not os.path.exists(git_exe):
            git_exe = shutil.which("git")

        if not git_exe or not os.path.exists(git_exe):
            return

        self.git_exe = git_exe

    @staticmethod
    def _git_is_real() -> bool:
        """On Macs, there is *always* an executable called "git", but sometimes it's just a
        script that tells the user to install XCode's Command Line tools. So the existence of git
        on the Mac actually requires us to check for that installation."""
        try:
            # Get the path to git from xcrun
            git_path = (
                subprocess.check_output(["xcrun", "--find", "git"], stderr=subprocess.DEVNULL)
                .decode()
                .strip()
            )
            if not os.path.exists(git_path) or not os.access(git_path, os.X_OK):
                return False

            # Check if running git triggers version output
            result = subprocess.run(
                [git_path, "--version"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
            )
            return result.returncode == 0

        except subprocess.CalledProcessError:
            return False
        except FileNotFoundError:
            return False

    def _synchronous_call_git(self, args: List[str]) -> str:
        """Calls git and returns its output."""
        final_args = [self.git_exe]
        final_args.extend(args)

        try:
            proc = utils.run_interruptable_subprocess(final_args)
        except subprocess.CalledProcessError as e:
            raise GitFailed(
                f"Git returned a non-zero exit status: {e.returncode}\n"
                + f"Called with: {' '.join(final_args)}\n\n"
                + f"Returned stderr:\n{e.stderr}"
            ) from e
        except utils.ProcessInterrupted as e:
            raise GitFailed(
                "The git process was interrupted due to a network timeout (or explicit user cancellation)\n"
                + f"Called with: {' '.join(final_args)}\n"
            ) from e

        return proc.stdout


def initialize_git() -> Optional[GitManager]:
    """If git is enabled, locate the git executable if necessary and return a new
    GitManager object. The executable location is saved in user preferences for reuse,
    and git can be disabled by setting the disableGit parameter in the Addons
    preference group. Returns None if for any of those reasons we aren't using git."""

    git_manager = None
    pref = fci.ParamGet("User parameter:BaseApp/Preferences/Addons")
    disable_git = pref.GetBool("disableGit", False)
    if not disable_git:
        try:
            git_manager = GitManager()
        except NoGitFound:
            # If git wasn't found, we call just fall through to returning the None that is already
            # assigned to git_manager
            pass
    return git_manager



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provide the Draft Workbench public programming interface.

The Draft module offers tools to create and manipulate 2D objects.
The functions in this file must be usable without requiring the
graphical user interface.
These functions can be used as the backend for the graphical commands
defined in `DraftTools.py`.
"""
## \addtogroup DRAFT
#  @{

import FreeCAD as App

if App.GuiUp:
    import Draft_rc

    gui = True
    # To prevent complaints from code checkers (flake8)
    True if Draft_rc.__name__ else False
else:
    gui = False

__title__ = "FreeCAD Draft Workbench"
__author__ = (
    "Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, " "Dmitry Chigrin, Daniel Falck"
)
__url__ = "https://www.freecad.org"

# ---------------------------------------------------------------------------
# Utility functions
# ---------------------------------------------------------------------------
from draftutils.utils import ARROW_TYPES as arrowtypes

from draftutils.utils import type_check, typecheck, precision, tolerance

from draftutils.utils import (
    get_real_name,
    getRealName,
    get_type,
    getType,
    get_objects_of_type,
    getObjectsOfType,
    is_clone,
    isClone,
    get_clone_base,
    getCloneBase,
    print_shape,
    printShape,
    compare_objects,
    compareObjects,
    shapify,
    is_closed_edge,
)

from draftutils.utils import (
    string_encode_coin,
    stringencodecoin,
    load_svg_patterns,
    loadSvgPatterns,
    svg_patterns,
    svgpatterns,
    get_rgb,
    getrgb,
    argb_to_rgba,
    rgba_to_argb,
    get_rgba_tuple,
)

from draftfunctions.svg import get_svg, getSVG

from draftfunctions.dxf import get_dxf, getDXF

from draftutils.gui_utils import (
    get3DView,
    get_3d_view,
    autogroup,
    removeHidden,
    remove_hidden,
    get_diffuse_color,
    apply_current_style,
    restore_view_object,
    formatObject,
    format_object,
    getSelection,
    get_selection,
    getSelectionEx,
    get_selection_ex,
    select,
    loadTexture,
    load_texture,
    get_bbox,
    end_all_events,
)

from draftutils.gui_utils import dim_symbol, dimSymbol, dim_dash, dimDash

from draftutils.groups import (
    is_group,
    get_group_names,
    getGroupNames,
    ungroup,
    get_windows,
    get_group_contents,
    getGroupContents,
    get_movable_children,
    getMovableChildren,
)

# ---------------------------------------------------------------------------
# Draft functions
# ---------------------------------------------------------------------------
from draftfunctions.array import array

from draftfunctions.cut import cut

from draftfunctions.downgrade import downgrade

from draftfunctions.draftify import draftify

from draftfunctions.extrude import extrude

from draftfunctions.fuse import fuse

from draftfunctions.heal import heal

from draftfunctions.move import move, move_vertex, move_edge, copy_moved_edge

from draftfunctions.rotate import rotate, rotate_vertex, rotate_edge, copy_rotated_edge

from draftfunctions.scale import scale, scale_vertex, scale_edge, copy_scaled_edge

from draftfunctions.join import join_wires, joinWires, join_two_wires, joinTwoWires

from draftfunctions.split import split

from draftfunctions.offset import offset

from draftfunctions.mirror import mirror

from draftfunctions.upgrade import upgrade


# ---------------------------------------------------------------------------
# Draft objects
# ---------------------------------------------------------------------------

# base object
from draftobjects.base import DraftObject, _DraftObject

# base viewprovider
from draftviewproviders.view_base import (
    ViewProviderDraft,
    _ViewProviderDraft,
    ViewProviderDraftAlt,
    _ViewProviderDraftAlt,
    ViewProviderDraftPart,
    _ViewProviderDraftPart,
)

# App::Link support, used by the arrays
from draftobjects.draftlink import DraftLink, _DraftLink
from draftviewproviders.view_draftlink import ViewProviderDraftLink, _ViewProviderDraftLink

# circle
from draftobjects.circle import Circle, _Circle
from draftmake.make_circle import make_circle, makeCircle

# arcs
from draftmake.make_arc_3points import make_arc_3points


# ellipse
from draftobjects.ellipse import Ellipse, _Ellipse
from draftmake.make_ellipse import make_ellipse, makeEllipse

# rectangle
from draftobjects.rectangle import Rectangle, _Rectangle
from draftmake.make_rectangle import make_rectangle, makeRectangle

if App.GuiUp:
    from draftviewproviders.view_rectangle import ViewProviderRectangle, _ViewProviderRectangle

# polygon
from draftobjects.polygon import Polygon, _Polygon
from draftmake.make_polygon import make_polygon, makePolygon

# wire and line
from draftobjects.wire import Wire, _Wire

from draftmake.make_line import make_line, makeLine
from draftmake.make_wire import make_wire, makeWire

if App.GuiUp:
    from draftviewproviders.view_wire import ViewProviderWire, _ViewProviderWire

# bspline
from draftobjects.bspline import BSpline, _BSpline
from draftmake.make_bspline import make_bspline, makeBSpline

if App.GuiUp:
    from draftviewproviders.view_bspline import ViewProviderBSpline, _ViewProviderBSpline

# bezcurve
from draftobjects.bezcurve import BezCurve, _BezCurve
from draftmake.make_bezcurve import make_bezcurve, makeBezCurve

if App.GuiUp:
    from draftviewproviders.view_bezcurve import ViewProviderBezCurve, _ViewProviderBezCurve

# copy
from draftmake.make_copy import make_copy
from draftmake.make_copy import make_copy as makeCopy

# clone
from draftobjects.clone import Clone, _Clone
from draftmake.make_clone import make_clone, clone

if App.GuiUp:
    from draftviewproviders.view_clone import ViewProviderClone, _ViewProviderClone

# point
from draftobjects.point import Point, _Point
from draftmake.make_point import make_point, makePoint

if App.GuiUp:
    from draftviewproviders.view_point import ViewProviderPoint, _ViewProviderPoint

# arrays
from draftobjects.array import Array, _Array
from draftmake.make_array import make_array, makeArray
from draftmake.make_orthoarray import (
    make_ortho_array,
    make_ortho_array2d,
    make_rect_array,
    make_rect_array2d,
)
from draftmake.make_polararray import make_polar_array
from draftmake.make_circulararray import make_circular_array

from draftobjects.patharray import PathArray, _PathArray
from draftmake.make_patharray import make_path_array, makePathArray, make_path_twisted_array

from draftobjects.pointarray import PointArray, _PointArray
from draftmake.make_pointarray import make_point_array, makePointArray

if App.GuiUp:
    from draftviewproviders.view_array import ViewProviderDraftArray, _ViewProviderDraftArray

# facebinder
from draftobjects.facebinder import Facebinder, _Facebinder
from draftmake.make_facebinder import make_facebinder, makeFacebinder

if App.GuiUp:
    from draftviewproviders.view_facebinder import ViewProviderFacebinder, _ViewProviderFacebinder

# shapestring
from draftobjects.block import Block, _Block
from draftmake.make_block import make_block, makeBlock

# shapestring
from draftobjects.shapestring import ShapeString, _ShapeString
from draftmake.make_shapestring import make_shapestring, makeShapeString

if App.GuiUp:
    from draftviewproviders.view_shapestring import ViewProviderShapeString

# shape 2d view
from draftobjects.shape2dview import Shape2DView, _Shape2DView
from draftmake.make_shape2dview import make_shape2dview, makeShape2DView

# sketch
from draftmake.make_sketch import make_sketch, makeSketch

# working plane proxy
from draftobjects.wpproxy import WorkingPlaneProxy
from draftmake.make_wpproxy import make_workingplaneproxy, makeWorkingPlaneProxy

if App.GuiUp:
    from draftviewproviders.view_wpproxy import ViewProviderWorkingPlaneProxy

from draftobjects.fillet import Fillet
from draftmake.make_fillet import make_fillet

if App.GuiUp:
    from draftviewproviders.view_fillet import ViewProviderFillet

from draftobjects.layer import Layer, _VisGroup, get_layer

from draftmake.make_layer import make_layer

if App.GuiUp:
    from draftviewproviders.view_layer import ViewProviderLayer, _ViewProviderVisGroup

# Annotation objects
from draftobjects.dimension import LinearDimension, _Dimension, AngularDimension, _AngularDimension
from draftmake.make_dimension import (
    make_dimension,
    makeDimension,
    make_linear_dimension,
    make_linear_dimension_obj,
    make_radial_dimension_obj,
    make_angular_dimension,
    makeAngularDimension,
)

if App.GuiUp:
    from draftviewproviders.view_dimension import (
        ViewProviderLinearDimension,
        _ViewProviderDimension,
        ViewProviderAngularDimension,
        _ViewProviderAngularDimension,
    )

from draftobjects.label import Label, DraftLabel
from draftmake.make_label import make_label, makeLabel

if App.GuiUp:
    from draftviewproviders.view_label import ViewProviderLabel, ViewProviderDraftLabel

from draftobjects.text import Text, DraftText
from draftmake.make_text import make_text, makeText, convert_draft_texts, convertDraftTexts

if App.GuiUp:
    from draftviewproviders.view_text import ViewProviderText, ViewProviderDraftText

from draftobjects.hatch import Hatch
from draftmake.make_hatch import make_hatch

if App.GuiUp:
    from draftviewproviders.view_hatch import ViewProviderDraftHatch

## @}



# ***************************************************************************
# *   Copyright (c) 2009 Juergen Riegel <juergen.riegel@web.de>             *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""FEM module Gui init script

Gathering all the information to start FreeCAD.
This is the second one of three init scripts.
The third one runs when the gui is up.

The script is executed using exec().
This happens inside srd/Gui/FreeCADGuiInit.py
All imports made there are available here too.
Thus no need to import them here.
But the import code line is used anyway to get flake8 quired.
Since they are cached they will not be imported twice.
"""

__title__ = "FEM module Gui init script"
__author__ = "Juergen Riegel, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

# imports to get flake8 quired
import sys
import FreeCAD
import FreeCADGui
from FreeCADGui import Workbench

# needed imports
from femguiutils.migrate_gui import FemMigrateGui


# migrate old FEM Gui objects
sys.meta_path.append(FemMigrateGui())


# add FEM Gui unit tests
# Disabled on 2025-10-30 because of unexplained failing CI that appears to trace back to these
# few tests. - chennes
# FreeCAD.__unit_test__ += ["TestFemGui"]


class FemWorkbench(Workbench):
    "Fem workbench object"

    def __init__(self):
        self.__class__.Icon = FreeCAD.getResourceDir() + "Mod/Fem/Resources/icons/FemWorkbench.svg"
        self.__class__.MenuText = "FEM"
        self.__class__.ToolTip = "FEM workbench"

    def Initialize(self):
        # load the module
        import Fem
        import FemGui
        import femcommands.commands
        import fempreferencepages

        FreeCADGui.addPreferencePage(fempreferencepages.DlgSettingsNetgen, "FEM")

        # dummy usage to get flake8 and lgtm quiet
        False if Fem.__name__ else True
        False if FemGui.__name__ else True
        False if femcommands.commands.__name__ else True

        # check vtk version to potentially find missmatchs
        if "BUILD_FEM_VTK_PYTHON" in FreeCAD.__cmake__:
            from femguiutils.vtk_module_handling import vtk_module_handling

            vtk_module_handling()

    def GetClassName(self):
        # see https://forum.freecad.org/viewtopic.php?f=10&t=43300
        return "FemGui::Workbench"


FreeCADGui.addWorkbench(FemWorkbench())
